import re

from loguru import logger
from sqlalchemy import desc

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session

from . import db

logger = logger.bind(name='next_series_episodes')


class NextSeriesEpisodes:
    """
    Emit next episode number from all series configured in this task.

    Supports only 'ep' and 'sequence' mode series.
    """

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'from_start': {'type': 'boolean', 'default': False},
                    'backfill': {'type': 'boolean', 'default': False},
                    'only_same_season': {'type': 'boolean', 'default': False},
                },
                'additionalProperties': False,
            },
        ]
    }

    def __init__(self):
        self.rerun_entries = []

    def ep_identifiers(self, season, episode):
        return ['S%02dE%02d' % (season, episode), '%dx%02d' % (season, episode)]

    def sequence_identifiers(self, episode):
        # Use a set to remove doubles, which will happen depending on number of digits in episode
        return {'%d' % episode, '%02d' % episode, '%03d' % episode}

    def search_entry(self, series, season, episode, task, rerun=True):
        # Extract the alternate names for the series
        alts = [alt.alt_name for alt in series.alternate_names]
        # Also consider series name without parenthetical (year, country) an alternate name
        paren_match = re.match(r'(.+?)( \(.+\))?$', series.name)
        if paren_match.group(2):
            alts.append(paren_match.group(1))
        if series.identified_by == 'ep':
            search_strings = [
                '%s %s' % (series.name, id) for id in self.ep_identifiers(season, episode)
            ]
            series_id = 'S%02dE%02d' % (season, episode)
            for alt in alts:
                search_strings.extend(
                    ['%s %s' % (alt, id) for id in self.ep_identifiers(season, episode)]
                )
        else:
            search_strings = [
                '%s %s' % (series.name, id) for id in self.sequence_identifiers(episode)
            ]
            series_id = episode
            for alt in alts:
                search_strings.extend(
                    ['%s %s' % (alt, id) for id in self.sequence_identifiers(episode)]
                )
        entry = Entry(
            title=search_strings[0],
            url='',
            search_strings=search_strings,
            series_name=series.name,
            series_alternate_names=alts,  # Not sure if this field is useful down the road.
            series_season=season,
            series_episode=episode,
            season_pack_lookup=False,
            series_id=series_id,
            series_id_type=series.identified_by,
        )
        if rerun:
            entry.on_complete(
                self.on_search_complete, task=task, identified_by=series.identified_by
            )
        return entry

    def on_task_input(self, task, config):
        if not config:
            return
        if isinstance(config, bool):
            config = {}
        self.config = config
        if task.is_rerun:
            # Just return calculated next eps on reruns
            entries = self.rerun_entries
            self.rerun_entries = []
            return entries
        else:
            self.rerun_entries = []

        entries = []
        impossible = {}
        with Session() as session:
            for seriestask in (
                session.query(db.SeriesTask).filter(db.SeriesTask.name == task.name).all()
            ):
                series = seriestask.series
                logger.trace('evaluating {}', series.name)
                if not series:
                    # TODO: How can this happen?
                    logger.debug('Found SeriesTask item without series specified. Cleaning up.')
                    session.delete(seriestask)
                    continue

                if series.identified_by not in ['ep', 'sequence']:
                    logger.trace('unsupported identified_by scheme')
                    reason = series.identified_by or 'auto'
                    impossible.setdefault(reason, []).append(series.name)
                    continue

                low_season = 0 if series.identified_by == 'ep' else -1
                # Don't look for seasons older than begin ep
                if series.begin and series.begin.season and series.begin.season > 1:
                    # begin-1 or the range() loop will never get to the begin season
                    low_season = max(series.begin.season - 1, 0)

                new_season = None
                check_downloaded = not config.get('backfill')
                latest_season = db.get_latest_release(series, downloaded=check_downloaded)
                if latest_season:
                    if latest_season.season <= low_season:
                        latest_season = new_season = low_season + 1
                    elif latest_season.season in series.completed_seasons:
                        latest_season = new_season = latest_season.season + 1
                    else:
                        latest_season = latest_season.season
                else:
                    latest_season = low_season + 1

                for season in range(latest_season, low_season, -1):
                    if season in series.completed_seasons:
                        logger.debug('season {} is marked as completed, skipping', season)
                        continue
                    logger.trace(
                        'Evaluating episodes for series {}, season {}', series.name, season
                    )
                    latest = db.get_latest_release(
                        series, season=season, downloaded=check_downloaded
                    )
                    if (
                        series.begin
                        and season == series.begin.season
                        and (not latest or latest < series.begin)
                    ):
                        # In case series.begin season is already completed, look in next available season
                        lookup_season = series.begin.season
                        ep_number = series.begin.number
                        while lookup_season in series.completed_seasons:
                            lookup_season += 1
                            # If season number was bumped, start looking for ep 1
                            ep_number = 1
                        entries.append(self.search_entry(series, lookup_season, ep_number, task))
                    elif latest and not config.get('backfill'):
                        entries.append(
                            self.search_entry(series, latest.season, latest.number + 1, task)
                        )
                    elif latest:
                        start_at_ep = 1
                        episodes_this_season = (
                            session.query(db.Episode)
                            .filter(db.Episode.series_id == series.id)
                            .filter(db.Episode.season == season)
                        )
                        if series.identified_by == 'sequence':
                            # Don't look for missing too far back with sequence shows
                            start_at_ep = max(latest.number - 10, 1)
                            episodes_this_season = episodes_this_season.filter(
                                db.Episode.number >= start_at_ep
                            )
                        latest_ep_this_season = episodes_this_season.order_by(
                            desc(db.Episode.number)
                        ).first()
                        if latest_ep_this_season:
                            downloaded_this_season = (
                                episodes_this_season.join(db.Episode.releases)
                                .filter(db.EpisodeRelease.downloaded == True)
                                .all()
                            )
                            # Calculate the episodes we still need to get from this season
                            if series.begin and series.begin.season == season:
                                start_at_ep = max(start_at_ep, series.begin.number)
                            eps_to_get = list(range(start_at_ep, latest_ep_this_season.number + 1))
                            for ep in downloaded_this_season:
                                try:
                                    eps_to_get.remove(ep.number)
                                except ValueError:
                                    pass
                            entries.extend(
                                self.search_entry(series, season, x, task, rerun=False)
                                for x in eps_to_get
                            )
                            # If we have already downloaded the latest known episode, try the next episode
                            if latest_ep_this_season.releases:
                                entries.append(
                                    self.search_entry(
                                        series, season, latest_ep_this_season.number + 1, task
                                    )
                                )
                        else:
                            # No episode means that latest is a season pack, emit episode 1
                            entries.append(self.search_entry(series, season, 1, task))
                    # First iteration of a new season with no show begin and show has downloads
                    elif new_season and season == new_season:
                        entries.append(self.search_entry(series, season, 1, task))
                    else:
                        if config.get('from_start') or config.get('backfill'):
                            entries.append(self.search_entry(series, season, 1, task))
                        else:
                            logger.verbose(
                                'Series `{}` has no history. Set begin option, or use CLI `series begin` '
                                'subcommand to set first episode to emit',
                                series.name,
                            )
                            break
                    # Skip older seasons if we are not in backfill mode
                    if not config.get('backfill'):
                        logger.debug('backfill is not enabled; skipping older seasons')
                        break

        for reason, series in impossible.items():
            logger.verbose(
                'Series `{}` with identified_by value `{}` are not supported. ',
                ', '.join(sorted(series)),
                reason,
            )

        return entries

    def on_search_complete(self, entry, task=None, identified_by=None, **kwargs):
        """Decides whether we should look for next ep/season based on whether we found/accepted any episodes."""
        with Session() as session:
            series = (
                session.query(db.Series).filter(db.Series.name == entry['series_name']).first()
            )
            latest = db.get_latest_release(series)
            db_release = (
                session.query(db.EpisodeRelease)
                .join(db.EpisodeRelease.episode)
                .join(db.Episode.series)
                .filter(db.Series.name == entry['series_name'])
                .filter(db.Episode.season == entry['series_season'])
                .filter(db.Episode.number == entry['series_episode'])
                .first()
            )

            if entry.accepted:
                logger.debug(
                    '{} {} was accepted, rerunning to look for next ep.',
                    entry['series_name'],
                    entry['series_id'],
                )
                self.rerun_entries.append(
                    self.search_entry(
                        series, entry['series_season'], entry['series_episode'] + 1, task
                    )
                )
                # Increase rerun limit by one if we have matches, this way
                # we keep searching as long as matches are found!
                # TODO: this should ideally be in discover so it would be more generic
                task.max_reruns += 1
                task.rerun(plugin='next_series_episodes', reason='Look for next episode')
            elif db_release:
                # There are know releases of this episode, but none were accepted
                return
            elif latest:
                if latest.is_season:
                    # A season pack was picked up in the task, no need to look for more episodes
                    return
                elif (
                    not self.config.get('only_same_season')
                    and identified_by == 'ep'
                    and (
                        entry['series_season'] == latest.season
                        and entry['series_episode'] == latest.number + 1
                    )
                ):
                    # We searched for next predicted episode of this season unsuccessfully, try the next season
                    self.rerun_entries.append(
                        self.search_entry(series, latest.season + 1, 1, task)
                    )
                    logger.debug(
                        '{} {} not found, rerunning to look for next season',
                        entry['series_name'],
                        entry['series_id'],
                    )
                    task.rerun(plugin='next_series_episodes', reason='Look for next season')


@event('plugin.register')
def register_plugin():
    plugin.register(NextSeriesEpisodes, 'next_series_episodes', api_ver=2)



import locale
import os
import os.path
import re
from contextlib import suppress
from copy import copy
from datetime import date, datetime, time
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Union, AnyStr, Type, cast

import jinja2.filters
from dateutil import parser as dateutil_parse
from jinja2 import (
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    StrictUndefined,
    Template,
    TemplateNotFound,
    TemplateSyntaxError,
)
from jinja2.nativetypes import NativeTemplate
from loguru import logger

from flexget.event import event
from flexget.utils.lazy_dict import LazyDict
from flexget.utils.pathscrub import pathscrub

if TYPE_CHECKING:
    from flexget.entry import Entry
    from flexget.manager import Manager
    from flexget.task import Task

logger = logger.bind(name='utils.template')

# The environment will be created after the manager has started
environment: Optional['FlexGetEnvironment'] = None


class RenderError(Exception):
    """Error raised when there is a problem with jinja rendering."""


def filter_pathbase(val: Optional[str]) -> str:
    """Base name of a path."""
    return os.path.basename(val or '')


def filter_pathname(val: Optional[str]) -> str:
    """Base name of a path, without its extension."""
    return os.path.splitext(os.path.basename(val or ''))[0]


def filter_pathext(val: Optional[str]) -> str:
    """Extension of a path (including the '.')."""
    return os.path.splitext(val or '')[1]


def filter_pathdir(val: Optional[str]) -> str:
    """Directory containing the given path."""
    return os.path.dirname(val or '')


def filter_pathscrub(val: str, os_mode: str = None) -> str:
    """Replace problematic characters in a path."""
    return pathscrub(val, os_mode)


def filter_re_replace(val: AnyStr, pattern: str, repl: str) -> str:
    """Perform a regexp replacement on the given string."""
    return re.sub(pattern, repl, str(val))


def filter_re_search(val, pattern: str):
    """Perform a search for given regexp pattern, return the matching portion of the text."""
    if not isinstance(val, str):
        return val
    result = re.search(pattern, val, re.IGNORECASE)
    if result:
        return result.group(0)
    return ''


def filter_formatdate(val, format_str):
    """Returns a string representation of a datetime object according to format string."""
    encoding = locale.getpreferredencoding()
    if not isinstance(val, (datetime, date, time)):
        return val
    return val.strftime(format_str)


def filter_parsedate(val):
    """Attempts to parse a date according to the rules in ISO 8601 and RFC 2822"""
    return dateutil_parse.parse(val)


def filter_date_suffix(date_str: str):
    """Returns a date suffix for a given date"""
    day = int(date_str[-2:])
    if 4 <= day <= 20 or 24 <= day <= 30:
        suffix = "th"
    else:
        suffix = ["st", "nd", "rd"][day % 10 - 1]
    return date_str + suffix


def filter_format_number(val, places: int = None, grouping: bool = True) -> str:
    """Formats a number according to the user's locale."""
    if not isinstance(val, (int, float)):
        return val
    if places is not None:
        format_str = f'%.{places}f'
    elif isinstance(val, int):
        format_str = '%d'
    else:
        format_str = '%.02f'

    locale.setlocale(locale.LC_ALL, '')
    return locale.format_string(format_str, val, grouping)


def filter_pad(val: Union[int, str], width: int, fillchar: str = '0') -> str:
    """Pads a number or string with fillchar to the specified width."""
    return str(val).rjust(width, fillchar)


def filter_to_date(date_time_val):
    """Returns the date from any date-time object"""
    if not isinstance(date_time_val, (datetime, date, time)):
        return date_time_val
    return date_time_val.date()


def filter_default(value, default_value: str = '', boolean: bool = True) -> str:
    """Override the built-in Jinja default filter to set the `boolean` param to True by default"""
    return jinja2.filters.do_default(value, default_value, boolean)


filter_d = filter_default


def is_fs_file(pathname: Union[str, os.PathLike]) -> bool:
    """Test whether item is existing file in filesystem"""
    return os.path.isfile(pathname)


def is_fs_dir(pathname: Union[str, os.PathLike]) -> bool:
    """Test whether item is existing directory in filesystem"""
    return os.path.isdir(pathname)


def is_fs_link(pathname: Union[str, os.PathLike]) -> bool:
    """Test whether item is existing link in filesystem"""
    return os.path.islink(pathname)


class FlexGetTemplate(Template):
    """Adds lazy lookup support when rendering templates."""

    def new_context(self, vars=None, shared=False, locals=None):
        context = super().new_context(vars, shared, locals)
        context.parent = LazyDict(context.parent)
        return context


class FlexGetNativeTemplate(FlexGetTemplate, NativeTemplate):
    """Lazy lookup support and native python return types."""


class FlexGetEnvironment(Environment):
    """Environment with template_class support"""

    template_class: Type[FlexGetTemplate]


@event('manager.initialize')
def make_environment(manager: 'Manager') -> None:
    """Create our environment and add our custom filters"""
    global environment
    environment = FlexGetEnvironment(
        undefined=StrictUndefined,
        loader=ChoiceLoader(
            [
                PackageLoader('flexget'),
                FileSystemLoader(os.path.join(manager.config_base, 'templates')),
            ]
        ),
        extensions=['jinja2.ext.loopcontrols'],
    )
    environment.template_class = FlexGetTemplate
    for name, filt in list(globals().items()):
        if name.startswith('filter_'):
            environment.filters[name.split('_', 1)[1]] = filt
    for name, test in list(globals().items()):
        if name.startswith('is_'):
            environment.tests[name.split('_', 1)[1]] = test


def list_templates(extensions: List[str] = None) -> List[str]:
    """Returns all templates names that are configured under environment loader dirs"""
    if environment is None or not hasattr(environment, 'loader'):
        return []
    return environment.list_templates(extensions=extensions)


def get_filters() -> dict:
    """Returns all built-in and custom Jinja filters in a dict

    The key is the name, and the value is the filter func
    """
    if environment is None or not hasattr(environment, 'loader'):
        return {}
    return environment.filters


def get_template(template_name: str, scope: Optional[str] = 'task') -> FlexGetTemplate:
    """Loads a template from disk. Looks in both included plugins and users custom scope dir."""

    if not template_name.endswith('.template'):
        template_name += '.template'
    locations = []
    if scope:
        locations.append(scope + '/' + template_name)
    locations.append(template_name)
    for location in locations:
        if environment is not None:
            with suppress(TemplateNotFound):
                return cast(FlexGetTemplate, environment.get_template(location))
    else:
        err = f'Template not found in templates dir: {template_name}'
        if scope:
            err += f' ({scope})'
        raise ValueError(err)


def render(template: Union[FlexGetTemplate, str], context: Mapping, native: bool = False) -> str:
    """
    Renders a Template with `context` as its context.

    :param template: Template or template string to render.
    :param context: Context to render the template from.
    :param native: If True, and the rendering result can be all native python types, not just strings.
    :return: The rendered template text.
    """
    if isinstance(template, str) and environment is not None:
        template_class = None
        if native:
            template_class = FlexGetNativeTemplate
        try:
            template = cast(
                FlexGetTemplate, environment.from_string(template, template_class=template_class)
            )
        except TemplateSyntaxError as e:
            raise RenderError(f'Error in template syntax: {e.message}')
    try:
        template = cast(FlexGetTemplate, template)
        result = template.render(context)
    except Exception as e:
        error = RenderError(f'({type(e).__name__}) {e}')
        logger.debug(f'Error during rendering: {error}')
        raise error

    return result


def render_from_entry(
    template: Union[FlexGetTemplate, str], entry: 'Entry', native: bool = False
) -> str:
    """Renders a Template or template string with an Entry as its context."""

    # Make a copy of the Entry so we can add some more fields
    variables = copy(entry.store)
    variables['now'] = datetime.now()
    # Add task name to variables, usually it's there because metainfo_task plugin, but not always
    if hasattr(entry, 'task') and entry.task is not None:
        if 'task' not in variables:
            variables['task'] = entry.task.name
        # Since `task` has different meaning between entry and task scope, the `task_name` field is create to be
        # consistent
        variables['task_name'] = entry.task.name
    return render(template, variables, native=native)


def render_from_task(template: Union[FlexGetTemplate, str], task: 'Task') -> str:
    """
    Renders a Template with a task as its context.

    :param template: Template or template string to render.
    :param task: Task to render the template from.
    :return: The rendered template text.
    """
    variables = {'task': task, 'now': datetime.now(), 'task_name': task.name}
    return render(template, variables)


def evaluate_expression(expression: str, context: Mapping) -> Any:
    """
    Evaluate a jinja `expression` using a given `context` with support for `LazyDict`s (`Entry`s.)

    :param str expression:  A jinja expression to evaluate
    :param context: dictlike, supporting LazyDicts
    """
    if environment is not None:
        compiled_expr = environment.compile_expression(expression)
        # If we have a LazyDict, grab the underlying store. Our environment supports LazyFields directly
        if isinstance(context, LazyDict):
            context = context.store
        return compiled_expr(**context)
    return None



import re
import unicodedata

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.components.sites.utils import normalize_unicode
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests
from flexget.utils.requests import Session, TimedLimiter
from flexget.utils.soup import get_soup

logger = logger.bind(name='descargas2020')

DESCARGAS2020_TORRENT_FORMAT = 'https://descargas2020.org/download/{:0>6}.torrent'
REWRITABLE_REGEX = re.compile(
    r'https?://(www.)?(descargas2020|tvsinpagar|tumejortorrent|torrentlocura|torrentrapid).(org|com)/'
)
NONREWRITABLE_REGEX = re.compile(r'(.*/descargar-torrent/|.*\.torrent$)')


class UrlRewriteDescargas2020:
    """Descargas2020 urlrewriter and search."""

    schema = {'type': 'boolean', 'default': False}

    def __init__(self):
        self._session = None

    @property
    def session(self):
        # TODO: This is not used for all requests even ..
        if self._session is None:
            self._session = Session()
            self._session.headers.update(
                {'User-Agent': 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)'}
            )
            self._session.add_domain_limiter(TimedLimiter('descargas2020.org', '2 seconds'))
        return self._session

    # urlrewriter API
    def url_rewritable(self, task, entry):
        url = entry['url']
        return REWRITABLE_REGEX.match(url) and not NONREWRITABLE_REGEX.match(url)

    # urlrewriter API
    def url_rewrite(self, task, entry):
        entry['url'] = self.parse_download_page(entry['url'], task)

    @plugin.internet(logger)
    def parse_download_page(self, url, task):
        logger.verbose('Descargas2020 URL: {}', url)

        try:
            page = self.session.get(url)
        except requests.RequestException as e:
            raise UrlRewritingError(e)
        try:
            soup = get_soup(page.text)
        except Exception as e:
            raise UrlRewritingError(e)

        torrent_id = None
        url_format = DESCARGAS2020_TORRENT_FORMAT

        torrent_id_prog = re.compile(
            r"(?:parametros\s*=\s*\n?)\s*{\s*\n(?:\s*'\w+'\s*:.*\n)+\s*'(?:torrentID|id)'\s*:\s*'(\d+)'"
        )
        torrent_ids = soup.findAll(text=torrent_id_prog)
        if torrent_ids:
            match = torrent_id_prog.search(torrent_ids[0])
            if match:
                torrent_id = match.group(1)
        if not torrent_id:
            logger.debug('torrent ID not found, searching openTorrent script')
            match = re.search(
                r'function openTorrent.*\n.*\{.*(\n.*)+window\.location\.href =\s*\".*\/(\d+.*)\";',
                page.text,
                re.MULTILINE,
            )
            if match:
                torrent_id = match.group(2).rstrip('/')

        if not torrent_id:
            raise UrlRewritingError('Unable to locate torrent ID from url %s' % url)

        return url_format.format(torrent_id)

    def search(self, task, entry, config=None):
        if not config:
            logger.debug('Descargas2020 disabled')
            return set()
        logger.debug('Search Descargas2020')
        url_search = 'https://descargas2020.org/buscar'
        results = set()
        for search_string in entry.get('search_strings', [entry['title']]):
            query = normalize_unicode(search_string)
            query = re.sub(r' \(\d\d\d\d\)$', '', query)
            logger.debug('Searching Descargas2020 {}', query)
            query = unicodedata.normalize('NFD', query).encode('ascii', 'ignore')
            data = {'q': query}
            try:
                response = task.requests.post(url_search, data=data)
            except requests.RequestException as e:
                logger.error('Error searching Descargas2020: {}', e)
                return results
            content = response.content
            soup = get_soup(content)
            soup2 = soup.find('ul', attrs={'class': 'buscar-list'})
            children = soup2.findAll('a', href=True)
            for child in children:
                entry = Entry()
                entry['url'] = child['href']
                entry_title = child.find('h2')
                if entry_title is None:
                    logger.debug('Ignore empty entry')
                    continue
                entry_title = entry_title.text
                if not entry_title:
                    continue
                try:
                    entry_quality_lan = re.search(
                        r'.+ \[([^\]]+)\](\[[^\]]+\])+$', entry_title
                    ).group(1)
                except AttributeError:
                    logger.debug('Quality not found')
                    continue
                entry_title = re.sub(r' \[.+]$', '', entry_title)
                entry['title'] = entry_title + ' ' + entry_quality_lan
                results.add(entry)
        logger.debug('Finish search Descargas2020 with {} entries', len(results))
        return results


@event('plugin.register')
def register_plugin():
    plugin.register(
        UrlRewriteDescargas2020, 'descargas2020', interfaces=['urlrewriter', 'search'], api_ver=2
    )



from flexget import options
from flexget.event import event
from flexget.manager import Session
from flexget.terminal import TerminalTable, TerminalTableError, console, table_parser

from . import db


def do_cli(manager, options):
    if options.failed_action == 'list':
        list_failed(options)
    elif options.failed_action == 'clear':
        clear_failed(manager)


def list_failed(options):
    with Session() as session:
        results = session.query(db.FailedEntry).all()
        header = ['#', 'Title', 'Fail count', 'Reason', 'Failure time']
        table_data = [header]
        for entry in results:
            table_data.append(
                [
                    entry.id,
                    entry.title,
                    entry.count,
                    '' if entry.reason == 'None' else entry.reason,
                    entry.tof.strftime('%Y-%m-%d %H:%M'),
                ]
            )
    try:
        table = TerminalTable(options.table_type, table_data, wrap_columns=[3, 1])
    except TerminalTableError as e:
        console('ERROR: %s' % str(e))
    else:
        table.table.justify_columns[0] = 'center'
        console(table.output)


def clear_failed(manager):
    # TODO: this should be a function in db.py
    with Session() as session:
        results = session.query(db.FailedEntry).delete()
        console('Cleared %i items.' % results)
        session.commit()
        if results:
            manager.config_changed()


@event('options.register')
def register_parser_arguments():
    parser = options.register_command('failed', do_cli, help='list or clear remembered failures')
    subparsers = parser.add_subparsers(dest='failed_action', metavar='<action>')
    subparsers.add_parser(
        'list', help='list all the entries that have had failures', parents=[table_parser]
    )
    subparsers.add_parser('clear', help='clear all failures from database, so they can be retried')



import re
from urllib.parse import urlparse, urlunparse

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.event import event
from flexget.utils.soup import get_soup

logger = logger.bind(name='eztv')

EZTV_MIRRORS = [('http', 'eztv.ch'), ('https', 'eztv-proxy.net'), ('http', 'eztv.come.in')]


class UrlRewriteEztv:
    """Eztv url rewriter."""

    def url_rewritable(self, task, entry):
        return urlparse(entry['url']).netloc == 'eztv.ch'

    def url_rewrite(self, task, entry):
        url = entry['url']
        page = None
        for (scheme, netloc) in EZTV_MIRRORS:
            try:
                _, _, path, params, query, fragment = urlparse(url)
                url = urlunparse((scheme, netloc, path, params, query, fragment))
                page = task.requests.get(url).content
            except RequestException as e:
                logger.debug('Eztv mirror `{}` seems to be down', url)
                continue
            break

        if not page:
            raise UrlRewritingError('No mirrors found for url %s' % entry['url'])

        logger.debug('Eztv mirror `{}` chosen', url)
        try:
            soup = get_soup(page)
            mirrors = soup.find_all('a', attrs={'class': re.compile(r'download_\d')})
        except Exception as e:
            raise UrlRewritingError(e)

        logger.debug('{} torrent mirrors found', len(mirrors))

        if not mirrors:
            raise UrlRewritingError('Unable to locate download link from url %s' % url)

        entry['urls'] = [m.get('href') for m in mirrors]
        entry['url'] = mirrors[0].get('href')


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteEztv, 'eztv', interfaces=['urlrewriter'], api_ver=2)



from http import HTTPStatus
from urllib.parse import urljoin

from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession

plugin_name = 'gotify'

requests = RequestSession(max_retries=3)


class GotifyNotifier(object):
    """
    Example::

    notify:
      entries:
        via:
          - gotify:
              url: <GOTIFY_SERVER_URL>
              token: <GOTIFY_TOKEN>
              priority: <PRIORITY>
    Configuration parameters are also supported from entries (eg. through set).
    """

    schema = {
        'type': 'object',
        'properties': {
            'url': {'format': 'url'},
            'token': {'type': 'string'},
            'priority': {'type': 'integer', 'default': 4},
        },
        'required': ['token', 'url'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send a Gotify notification
        """
        base_url = config['url']
        api_endpoint = '/message'
        url = urljoin(base_url, api_endpoint)
        params = {'token': config['token']}

        priority = config['priority']

        notification = {'title': title, 'message': message, 'priority': priority}
        # Make the request
        try:
            response = requests.post(url, params=params, json=notification)
        except RequestException as e:
            if e.response is not None:
                if e.response.status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
                    message = 'Invalid Gotify access token'
                else:
                    message = e.response.json()['error']['message']
            else:
                message = str(e)
            raise PluginWarning(message)


@event('plugin.register')
def register_plugin():
    plugin.register(GotifyNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'join'
logger = logger.bind(name=plugin_name)

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('appspot.com', '5 seconds'))

JOIN_URL = 'https://joinjoaomgcd.appspot.com/_ah/api/messaging/v1/sendPush'


class JoinNotifier:
    """
    Example::

      notify:
        entries:
          via:
            - join:
                [api_key: <API_KEY> (your join api key. Only required for 'group' notifications)]
                [group: <GROUP_NAME> (name of group of join devices to notify. 'all', 'android', etc.)
                [device: <DEVICE_ID> (can also be a list of device ids)]
                [url: <NOTIFICATION_URL>]
                [sms_number: <NOTIFICATION_SMS_NUMBER>]
                [icon: <NOTIFICATION_ICON>]
    """

    schema = {
        'type': 'object',
        'properties': {
            'api_key': {'type': 'string'},
            'group': {
                'type': 'string',
                'enum': ['all', 'android', 'chrome', 'windows10', 'phone', 'tablet', 'pc'],
            },
            'device': one_or_more({'type': 'string'}),
            'device_name': one_or_more({'type': 'string'}),
            'url': {'type': 'string'},
            'icon': {'type': 'string'},
            'sms_number': {'type': 'string'},
            'priority': {'type': 'integer', 'minimum': -2, 'maximum': 2},
        },
        'required': ['api_key'],
        'not': {'required': ['device', 'group']},
        'error_not': 'Cannot select both \'device\' and \'group\'',
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send Join notifications.
        """
        notification = {
            'title': title,
            'text': message,
            'url': config.get('url'),
            'icon': config.get('icon'),
            'priority': config.get('priority'),
            'apikey': config['api_key'],
        }
        if config.get('device'):
            if isinstance(config['device'], list):
                notification['deviceIds'] = ','.join(config['device'])
            else:
                notification['deviceId'] = config['device']
        elif config.get('group'):
            notification['deviceId'] = 'group.' + config['group']
        else:
            notification['deviceId'] = 'group.all'

        if config.get('device_name'):
            if isinstance(config['device_name'], list):
                notification['deviceNames'] = ','.join(config['device_name'])
            else:
                notification['deviceNames'] = config['device_name']

        if config.get('sms_number'):
            notification['smsnumber'] = config['sms_number']
            notification['smstext'] = message

        try:
            response = requests.get(JOIN_URL, params=notification)
        except RequestException as e:
            raise PluginWarning(e.args[0])
        else:
            error = response.json().get('errorMessage')
            if error:
                raise PluginWarning(error)


@event('plugin.register')
def register_plugin():
    plugin.register(JoinNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.components.sites.utils import normalize_scene
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='argenteam')


class SearchArgenteam:
    """Argenteam
    Search plugin which gives results from www.argenteam.net, latin american (Argentina) web.

    Configuration:
      - force_subtitles: [yes/no] #Force download release with subtitles made by aRGENTeaM. Default is yes

    Example
      argenteam:
        force_subtitles: yes
    """

    schema = {
        'type': 'object',
        'properties': {'force_subtitles': {'type': 'boolean', 'default': True}},
        "additionalProperties": False,
    }

    base_url = 'http://www.argenteam.net/api/v1/'

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for releases
        """

        entries = set()

        for search_string in entry.get('search_strings', [entry['title']]):

            try:
                params = {'q': normalize_scene(search_string)}
                resp = task.requests.get(self.base_url + 'search', params=params)
                logger.debug('Requesting: {}', resp.url)
                response = resp.json()
            except RequestException as e:
                logger.error('Argenteam request failed: {}', e)
                return

            if not response:
                logger.debug('Empty response from Argenteam')
                continue

            if not response.get('total'):
                logger.debug('No results found for {}', search_string)
                continue

            results = response.get('results')
            if results[0]['type'] == 'tvshow':
                logger.error('Argenteam type tvshow not supported yet.')
                continue

            url = '{}{}?id={}'.format(self.base_url, results[0]['type'], results[0]['id'])
            try:
                resp = task.requests.get(url)
                logger.debug('Requesting releases for: {}', url)
                response = resp.json()
            except RequestException as e:
                logger.error('Argenteam request failed: {}', e)
                return

            logger.debug('{} releases found.', len(response['releases']))
            for release in response['releases']:
                for torrent in release['torrents']:
                    if (
                        config.get('force_subtitles')
                        and release['subtitles']
                        or not config.get('force_subtitles')
                    ):
                        e = Entry()

                        e['title'] = ' '.join(
                            (
                                search_string,
                                release['source'],
                                release['codec'],
                                release['team'],
                                release['tags'],
                            )
                        )
                        e['url'] = torrent['uri']

                        # Save aRGENTeaM subtitle URL for this release
                        if 'subtitles' in release and len(release['subtitles']) > 0:
                            e['argenteam_subtitle'] = release['subtitles'][0]['uri']
                            logger.debug('Argenteam subtitle found: {}', e['argenteam_subtitle'])

                        if 'tvdb' in response:
                            e['tvdb_id'] = response['tvdb']
                        if 'info' in response and 'imdb' in response['info']:
                            e['imdb_id'] = response['info']['imdb']

                        entries.add(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchArgenteam, 'argenteam', interfaces=['search'], api_ver=2)



import hashlib

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'sms_ru'
logger = logger.bind(name=plugin_name)

SMS_SEND_URL = 'http://sms.ru/sms/send'
SMS_TOKEN_URL = 'http://sms.ru/auth/get_token'

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('sms.ru', '5 seconds'))


class SMSRuNotifier:
    """
    Sends SMS notification through sms.ru http api sms/send.
    Phone number is a login assigned to sms.ru account.

    Example:

      notify:
        entries:
          via:
            - sms_ru:
                phone_number: <PHONE_NUMBER> (accepted format example: '79997776655')
                password: <PASSWORD>

    """

    schema = {
        'type': 'object',
        'properties': {'phone_number': {'type': 'string'}, 'password': {'type': 'string'}},
        'additionalProperties': False,
        'required': ['phone_number', 'password'],
    }

    def notify(self, title, message, config):
        """
        Send an SMS RU notification
        """
        try:
            token_response = requests.get(SMS_TOKEN_URL)
        except RequestException as e:
            raise PluginWarning('Could not get auth token: %s' % repr(e))

        sha512 = hashlib.sha512(config['password'] + token_response.text).hexdigest()

        # Build request params
        notification = {
            'login': config['phone_number'],
            'sha512': sha512,
            'token': token_response.text,
            'to': config['phone_number'],
            'text': message,
        }

        try:
            response = requests.get(SMS_SEND_URL, params=notification)
        except RequestException as e:
            raise PluginWarning(e.args[0])
        else:
            if not response.text.find('100') == 0:
                raise PluginWarning(response.text)


@event('plugin.register')
def register_plugin():
    plugin.register(SMSRuNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='nzbget')


class OutputNzbget:
    """
    Example::

      nzbget:
        url: http://nzbget:12345@localhost:6789/xmlrpc
        category: movies
        priority: 0
        top: False
    """

    schema = {
        'type': 'object',
        'properties': {
            'url': {'type': 'string'},
            'category': {'type': 'string', 'default': ''},
            'priority': {'type': 'integer', 'default': 0},
            'top': {'type': 'boolean', 'default': False},
        },
        'required': ['url'],
        'additionalProperties': False,
    }

    def on_task_output(self, task, config):
        from xmlrpc.client import ServerProxy

        params = dict(config)

        server = ServerProxy(params["url"])

        for entry in task.accepted:
            if task.options.test:
                logger.info('Would add into nzbget: {}', entry['title'])
                continue

            # allow overriding the category
            if 'category' in entry:
                params['category'] = entry['category']

            try:
                server.appendurl(
                    entry['title'] + '.nzb',
                    params['category'],
                    params['priority'],
                    params['top'],
                    entry['url'],
                )
                logger.info('Added `{}` to nzbget', entry['title'])
            except Exception as e:
                logger.critical('rpc call to nzbget failed: {}', e)
                entry.fail('could not call appendurl via RPC')


@event('plugin.register')
def register_plugin():
    plugin.register(OutputNzbget, 'nzbget', api_ver=2)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.utils.template import RenderError

logger = logger.bind(name='set')

# Use a string for this sentinel, so it survives serialization
UNSET = '__unset__'


class ModifySet:
    """Allows adding information to a task entry for use later.

    Example:

    set:
      path: ~/download/path/
    """

    schema = {'type': 'object', "minProperties": 1}

    def on_task_metainfo(self, task, config):
        """Adds the set dict to all accepted entries."""
        for entry in task.all_entries:
            self.modify(entry, config)

    def modify(self, entry, config, errors=True):
        """This can be called from a plugin to add set values to an entry"""
        for field in config:
            # If this doesn't appear to be a jinja template, just set it right away.
            if not isinstance(config[field], str) or '{' not in config[field]:
                entry[field] = config[field]
            # Store original values before overwriting with a lazy field, so that set directives can reference
            # themselves.
            else:
                orig_value = entry.get(field, UNSET, eval_lazy=False)
                try:
                    del entry[field]
                except KeyError:
                    pass
                entry.add_lazy_fields(
                    self.lazy_set,
                    [field],
                    kwargs={
                        'config': config,
                        'field': field,
                        'orig_field_value': orig_value,
                        'errors': errors,
                    },
                )

    @entry.register_lazy_lookup('set_field')
    def lazy_set(self, entry, config, field, orig_field_value, errors=True):
        level = 'ERROR' if errors else 'DEBUG'
        if orig_field_value is not UNSET:
            entry[field] = orig_field_value
        try:
            entry[field] = entry.render(config[field], native=True)
        except RenderError as e:
            logger.log(level, 'Could not set {} for {}: {}', field, entry['title'], e)


@event('plugin.register')
def register_plugin():
    plugin.register(ModifySet, 'set', api_ver=2)



from loguru import logger
from sqlalchemy.orm.exc import NoResultFound

from flexget import plugin
from flexget.event import event
from flexget.manager import Session

from . import db

logger = logger.bind(name=__name__)


class EntryList:
    schema = {'type': 'string'}

    @staticmethod
    def get_list(config):
        return db.DBEntrySet(config)

    def on_task_input(self, task, config):
        return list(db.DBEntrySet(config))

    def search(self, task, entry, config=None):
        entries = []
        with Session() as session:
            try:
                entry_list = db.get_list_by_exact_name(config, session=session)
            except NoResultFound:
                logger.warning("Entry list with name '{}' does not exist", config)
            else:
                for search_string in entry.get('search_strings', [entry['title']]):
                    logger.debug(
                        'searching for entry that matches {} in entry_list {}',
                        search_string,
                        config,
                    )
                    search_string = search_string.replace(' ', '%').replace('.', '%')
                    query = entry_list.entries.filter(
                        db.EntryListEntry.title.like('%' + search_string + '%')
                    )
                    entries += [e.entry for e in query.all()]
            finally:
                return entries


@event('plugin.register')
def register_plugin():
    plugin.register(EntryList, 'entry_list', api_ver=2, interfaces=['task', 'list', 'search'])



from urllib.parse import quote

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils import json
from flexget.utils.template import RenderError

logger = logger.bind(name='pyload')


class PyloadApi:
    def __init__(self, requests, url):
        self.requests = requests
        self.url = url

    def get_session(self, config):
        # Login
        data = {'username': config['username'], 'password': config['password']}
        result = self.post('login', data=data)
        response = result.json()
        if not response:
            raise plugin.PluginError('Login failed', logger)

        if isinstance(response, str):
            return response.replace('"', '')
        else:
            return response

    def get(self, method):
        try:
            return self.requests.get(self.url.rstrip("/") + "/" + method.strip("/"))
        except RequestException as e:
            if e.response and e.response.status_code == 500:
                raise plugin.PluginError(
                    'Internal API Error: <%s> <%s>' % (method, self.url), logger
                )
            raise

    def post(self, method, data):
        try:
            return self.requests.post(self.url.rstrip("/") + "/" + method.strip("/"), data=data)
        except RequestException as e:
            if e.response and e.response.status_code == 500:
                raise plugin.PluginError(
                    'Internal API Error: <%s> <%s> <%s>' % (method, self.url, data), logger
                )
            raise


class PluginPyLoad:
    """
    Parse task content or url for hoster links and adds them to pyLoad.

    Example::

      pyload:
        api: http://localhost:8000/api
        queue: yes
        username: my_username
        password: my_password
        folder: desired_folder
        package: desired_package_name (jinja2 supported)
        package_password: desired_package_password
        hoster:
          - YoutubeCom
        parse_url: no
        multiple_hoster: yes
        enabled: yes

    Default values for the config elements::

      pyload:
          api: http://localhost:8000/api
          queue: no
          hoster: ALL
          parse_url: no
          multiple_hoster: yes
          enabled: yes
    """

    __author__ = 'http://pyload.org'
    __version__ = '0.5'

    DEFAULT_API = 'http://localhost:8000/api'
    DEFAULT_QUEUE = False
    DEFAULT_FOLDER = ''
    DEFAULT_HOSTER = []
    DEFAULT_PARSE_URL = False
    DEFAULT_MULTIPLE_HOSTER = True
    DEFAULT_PREFERRED_HOSTER_ONLY = False
    DEFAULT_HANDLE_NO_URL_AS_FAILURE = False

    schema = {
        'type': 'object',
        'properties': {
            'api': {'type': 'string'},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'folder': {'type': 'string'},
            'package': {'type': 'string'},
            'package_password': {'type': 'string'},
            'queue': {'type': 'boolean'},
            'parse_url': {'type': 'boolean'},
            'multiple_hoster': {'type': 'boolean'},
            'hoster': one_or_more({'type': 'string'}),
            'preferred_hoster_only': {'type': 'boolean'},
            'handle_no_url_as_failure': {'type': 'boolean'},
            'enabled': {'type': 'boolean'},
        },
        'required': ['username', 'password'],
        'additionalProperties': False,
    }

    def on_task_output(self, task, config):
        if not config.get('enabled', True):
            return
        if not task.accepted:
            return

        self.add_entries(task, config)

    @staticmethod
    def get_version_from_packaging():
        version = None
        try:
            from packaging import version
        except ModuleNotFoundError:
            logger.warning('packaging is not installed')
        return version

    def add_entries(self, task, config):
        """Adds accepted entries"""

        apiurl = config.get('api', self.DEFAULT_API)
        api = PyloadApi(task.requests, apiurl)

        try:
            session = api.get_session(config)
        except OSError:
            raise plugin.PluginError('pyLoad not reachable', logger)
        except plugin.PluginError:
            raise
        except Exception as e:
            raise plugin.PluginError('Unknown error: %s' % str(e), logger)

        remote_version = None
        try:
            remote_version = api.get('getServerVersion')
        except RequestException as e:
            if e.response is not None and e.response.status_code == 404:
                remote_version = json.loads(api.get('get_server_version').content)
            else:
                raise e

        parse_urls_command = 'parseURLs'
        add_package_command = 'addPackage'
        set_package_data_command = 'setPackageData'

        is_pyload_ng = False
        version = self.get_version_from_packaging()
        if version and version.parse(remote_version) >= version.parse('0.5'):
            parse_urls_command = 'parse_urls'
            add_package_command = 'add_package'
            set_package_data_command = 'set_package_date'
            is_pyload_ng = True

        hoster = config.get('hoster', self.DEFAULT_HOSTER)

        for entry in task.accepted:
            # bunch of urls now going to check
            content = entry.get('description', '') + ' ' + quote(entry['url'])
            content = json.dumps(content)

            if is_pyload_ng:
                url = entry['url'] if config.get('parse_url', self.DEFAULT_PARSE_URL) else ''
            else:
                url = (
                    json.dumps(entry['url'])
                    if config.get('parse_url', self.DEFAULT_PARSE_URL)
                    else "''"
                )

            logger.debug('Parsing url {}', url)

            data = {'html': content, 'url': url}
            if not is_pyload_ng:
                data['session'] = session
            result = api.post(parse_urls_command, data=data)

            parsed = result.json()

            urls = []

            # check for preferred hoster
            for name in hoster:
                if name in parsed:
                    urls.extend(parsed[name])
                    if not config.get('multiple_hoster', self.DEFAULT_MULTIPLE_HOSTER):
                        break

            # no preferred hoster and not preferred hoster only - add all recognized plugins
            if not urls and not config.get(
                'preferred_hoster_only', self.DEFAULT_PREFERRED_HOSTER_ONLY
            ):
                for name, purls in parsed.items():
                    if name != 'BasePlugin':
                        urls.extend(purls)

            if task.options.test:
                logger.info('Would add `{}` to pyload', urls)
                continue

            # no urls found
            if not urls:
                if config.get('handle_no_url_as_failure', self.DEFAULT_HANDLE_NO_URL_AS_FAILURE):
                    entry.fail('No suited urls in entry %s' % entry['title'])
                else:
                    logger.info('No suited urls in entry {}', entry['title'])
                continue

            logger.debug('Add {} urls to pyLoad', len(urls))

            try:
                dest = 1 if config.get('queue', self.DEFAULT_QUEUE) else 0  # Destination.Queue = 1

                # Use the title of the entry, if no naming schema for the package is defined.
                name = config.get('package', entry['title'])

                # If name has jinja template, render it
                try:
                    name = entry.render(name)
                except RenderError as e:
                    name = entry['title']
                    logger.error('Error rendering jinja event: {}', e)

                if is_pyload_ng:
                    data = {
                        'name': name.encode('ascii', 'ignore').decode(),
                        'links': urls,
                        'dest': dest,
                    }
                else:
                    data = {
                        'name': json.dumps(name.encode('ascii', 'ignore').decode()),
                        'links': json.dumps(urls),
                        'dest': json.dumps(dest),
                        'session': session,
                    }

                pid = api.post(add_package_command, data=data).text
                logger.debug('added package pid: {}', pid)

                # Set Folder
                folder = config.get('folder', self.DEFAULT_FOLDER)
                folder = entry.get('path', folder)
                if folder:
                    # If folder has jinja template, render it
                    try:
                        folder = entry.render(folder)
                    except RenderError as e:
                        folder = self.DEFAULT_FOLDER
                        logger.error('Error rendering jinja event: {}', e)
                    # set folder with api
                    data = json.dumps({'folder': folder})
                    post_data = {'pid': pid, 'data': data}
                    if not is_pyload_ng:
                        post_data['session'] = session
                    api.post(set_package_data_command, data=post_data)

                # Set Package Password
                package_password = config.get('package_password')
                if package_password:
                    data = json.dumps({'password': package_password})
                    post_data = {'pid': pid, 'data': data}
                    if not is_pyload_ng:
                        post_data['session'] = session
                    api.post(set_package_data_command, data=post_data)

            except Exception as e:
                entry.fail(str(e))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginPyLoad, 'pyload', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached
from flexget.utils.requests import RequestException

logger = logger.bind(name='my_anime_list')

STATUS = {'watching': 1, 'completed': 2, 'on_hold': 3, 'dropped': 4, 'plan_to_watch': 6, 'all': 7}

AIRING_STATUS = {'airing': 1, 'finished': 2, 'planned': 3, 'all': 6}

ANIME_TYPE = ['all', 'tv', 'ova', 'movie', 'special', 'ona', 'music', 'unknown']


class MyAnimeList:
    """ " Creates entries for series and movies from MyAnimeList list

    Syntax:
    my_anime_list:
      username: <value>
      status:
        - <watching|completed|on_hold|dropped|plan_to_watch>
        - <watching|completed|on_hold|dropped|plan_to_watch>
        ...
      airing_status:
        - <airing|finished|planned>
        - <airing|finished|planned>
        ...
      type:
        - <series|ova...>
    """

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'status': one_or_more(
                {'type': 'string', 'enum': list(STATUS.keys()), 'default': 'all'},
                unique_items=True,
            ),
            'airing_status': one_or_more(
                {'type': 'string', 'enum': list(AIRING_STATUS.keys()), 'default': 'all'},
                unique_items=True,
            ),
            'type': one_or_more(
                {'type': 'string', 'enum': list(ANIME_TYPE), 'default': 'all'}, unique_items=True
            ),
        },
        'required': ['username'],
        'additionalProperties': False,
    }

    @cached('my_anime_list', persist='2 hours')
    def on_task_input(self, task, config):
        entries = []
        selected_status = config['status']
        selected_airing_status = config['airing_status']
        selected_types = config['type']

        if not isinstance(selected_status, list):
            selected_status = [selected_status]

        if not isinstance(selected_airing_status, list):
            selected_airing_status = [selected_airing_status]

        if not isinstance(selected_types, list):
            selected_types = [selected_types]

        selected_status = [STATUS[s] for s in selected_status]

        selected_airing_status = [AIRING_STATUS[s] for s in selected_airing_status]

        try:
            list_response = task.requests.get(
                'https://myanimelist.net/animelist/' + config['username'] + '/load.json'
            )
        except RequestException as e:
            raise plugin.PluginError('Error finding list on url: {url}'.format(url=e.request.url))

        try:
            list_json = list_response.json()
        except ValueError:
            raise plugin.PluginError('Invalid JSON response')

        for anime in list_json:
            has_selected_status = anime["status"] in selected_status or config['status'] == 'all'
            has_selected_airing_status = (
                anime["anime_airing_status"] in selected_airing_status
                or config['airing_status'] == 'all'
            )
            has_selected_type = (
                anime["anime_media_type_string"].lower() in selected_types
                or config['type'] == 'all'
            )
            if has_selected_status and has_selected_type and has_selected_airing_status:
                # MAL sometimes returns title as an integer :| GH #2901
                anime["anime_title"] = str(anime["anime_title"])
                entries.append(
                    Entry(
                        title=anime["anime_title"],
                        url="https://myanimelist.net" + anime["anime_url"],
                        mal_name=anime["anime_title"],
                        mal_poster=anime["anime_image_path"],
                        mal_type=anime["anime_media_type_string"],
                        mal_tags=anime["tags"],
                    )
                )
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(MyAnimeList, 'my_anime_list', api_ver=2)



import logging
import os
import tempfile

from loguru import logger

from flexget import entry, plugin
from flexget.event import event

logger = logger.bind(name='check_subtitles')


class MetainfoSubs:
    """
    Set 'subtitles' field for entries, if they are local video files with subs.
    The field is a list of language codes (3-letter ISO-639-3) for each subtitles
    file found on disk and/or subs track found inside video (for MKVs).
    Special "und" code is for unidentified language (i.e. files without language
    code before extension).
    """

    schema = {'type': 'boolean'}

    def on_task_start(self, task, config):
        try:
            import subliminal
        except ImportError as e:
            logger.debug('Error importing Subliminal: {}', e)
            raise plugin.DependencyError(
                'subliminal', 'subliminal', 'Subliminal module required. ImportError: %s' % e
            )
        from dogpile.cache.exception import RegionAlreadyConfigured
        from subliminal.cli import MutexLock

        try:
            subliminal.region.configure(
                'dogpile.cache.dbm',
                arguments={
                    'filename': os.path.join(tempfile.gettempdir(), 'cachefile.dbm'),
                    'lock_factory': MutexLock,
                },
            )
        except RegionAlreadyConfigured:
            pass
        logging.getLogger("subliminal").setLevel(logging.CRITICAL)
        logging.getLogger("enzyme").setLevel(logging.WARNING)

    def on_task_metainfo(self, task, config):
        # check if explicitly disabled (value set to false)
        if config is False:
            return
        for entry in task.entries:
            entry.add_lazy_fields(self.get_subtitles, ['subtitles'])

    @entry.register_lazy_lookup('subtitles_check')
    def get_subtitles(self, entry):
        if (
            entry.get('subtitles', eval_lazy=False)
            or not ('location' in entry)
            or ('$RECYCLE.BIN' in entry['location'])
            or not os.path.exists(entry['location'])
        ):
            return
        from subliminal import scan_video
        from subliminal.core import refine, search_external_subtitles

        try:
            video = scan_video(entry['location'])
            # grab external and internal subtitles
            subtitles = video.subtitle_languages
            refiner = ('metadata',)
            refine(video, episode_refiners=refiner, movie_refiners=refiner)
            subtitles |= set(search_external_subtitles(entry['location']).values())
            if subtitles:
                # convert to human-readable strings
                subtitles = [str(l) for l in subtitles]
                entry['subtitles'] = subtitles
                logger.debug('Found subtitles {} for {}', '/'.join(subtitles), entry['title'])
        except Exception as e:
            logger.error('Error checking local subtitles for {}: {}', entry['title'], e)


@event('plugin.register')
def register_plugin():
    plugin.register(MetainfoSubs, 'check_subtitles', api_ver=2)



import re
from urllib.parse import quote

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.components.sites.utils import normalize_unicode, torrent_availability
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests
from flexget.utils.soup import get_soup

logger = logger.bind(name='newtorrents')


class NewTorrents:
    """NewTorrents urlrewriter and search plugin."""

    def __init__(self):
        self.resolved = []

    # UrlRewriter plugin API
    def url_rewritable(self, task, entry):
        # Return true only for urls that can and should be resolved
        if entry['url'].startswith('http://www.newtorrents.info/down.php?'):
            return False
        return (
            entry['url'].startswith('http://www.newtorrents.info')
            and not entry['url'] in self.resolved
        )

    # UrlRewriter plugin API
    def url_rewrite(self, task, entry):
        url = entry['url']
        if url.startswith('http://www.newtorrents.info/?q=') or url.startswith(
            'http://www.newtorrents.info/search'
        ):
            results = self.entries_from_search(entry['title'], url=url)
            if not results:
                raise UrlRewritingError("No matches for %s" % entry['title'])
            url = results[0]['url']
        else:
            url = self.url_from_page(url)

        if url:
            entry['url'] = url
            self.resolved.append(url)
        else:
            raise UrlRewritingError('Bug in newtorrents urlrewriter')

    # Search plugin API
    def search(self, task, entry, config=None):
        entries = set()
        for search_string in entry.get('search_string', [entry['title']]):
            entries.update(self.entries_from_search(search_string))
        return entries

    @plugin.internet(logger)
    def url_from_page(self, url):
        """Parses torrent url from newtorrents download page"""
        try:
            page = requests.get(url)
            data = page.text
        except requests.RequestException:
            raise UrlRewritingError('URLerror when retrieving page')
        p = re.compile(r"copy\(\'(.*)\'\)", re.IGNORECASE)
        f = p.search(data)
        if not f:
            # the link in which plugin relies is missing!
            raise UrlRewritingError(
                'Failed to get url from download page. Plugin may need a update.'
            )
        else:
            return f.group(1)

    @plugin.internet(logger)
    def entries_from_search(self, name, url=None):
        """Parses torrent download url from search results"""
        name = normalize_unicode(name)
        if not url:
            url = 'http://www.newtorrents.info/search/%s' % quote(
                name.encode('utf-8'), safe=b':/~?=&%'
            )

        logger.debug('search url: {}', url)

        html = requests.get(url).text
        # fix </SCR'+'IPT> so that BS does not crash
        # TODO: should use beautifulsoup massage
        html = re.sub(r'(</SCR.*?)...(.*?IPT>)', r'\1\2', html)

        soup = get_soup(html)
        # saving torrents in dict
        torrents = []
        for link in soup.find_all('a', attrs={'href': re.compile('down.php')}):
            torrent_url = 'http://www.newtorrents.info%s' % link.get('href')
            release_name = link.parent.next.get('title')
            # quick dirty hack
            seed = link.find_next('td', attrs={'class': re.compile('s')}).renderContents()
            if seed == 'n/a':
                seed = 0
            else:
                try:
                    seed = int(seed)
                except ValueError:
                    logger.warning(
                        'Error converting seed value ({}) from newtorrents to integer.', seed
                    )
                    seed = 0

            # TODO: also parse content_size and peers from results
            torrents.append(
                Entry(
                    title=release_name,
                    url=torrent_url,
                    torrent_seeds=seed,
                    torrent_availability=torrent_availability(seed, 0),
                )
            )
        # sort with seed number Reverse order
        torrents.sort(reverse=True, key=lambda x: x.get('torrent_availability', 0))
        # choose the torrent
        if not torrents:
            dashindex = name.rfind('-')
            if dashindex != -1:
                return self.entries_from_search(name[:dashindex])
            else:
                return torrents
        else:
            if len(torrents) == 1:
                logger.debug('found only one matching search result.')
            else:
                logger.debug(
                    'search result contains multiple matches, sorted {} by most seeders', torrents
                )
            return torrents


@event('plugin.register')
def register_plugin():
    plugin.register(NewTorrents, 'newtorrents', interfaces=['urlrewriter', 'search'], api_ver=2)



from collections import defaultdict

from loguru import logger

from flexget import plugin
from flexget.event import event

from . import seen as plugin_seen

logger = logger.bind(name='seen_movies')


class FilterSeenMovies(plugin_seen.FilterSeen):
    """
    Prevents movies being downloaded twice.
    Works only on entries which have imdb url available.

    How duplicate movie detection works:
    1) Remember all imdb urls from downloaded entries.
    2) If stored imdb url appears again, entry is rejected.
    """

    schema = {
        'oneOf': [
            {'type': 'string', 'enum': ['strict', 'loose']},
            {
                'type': 'object',
                'properties': {
                    'scope': {'type': 'string', 'enum': ['global', 'local']},
                    'matching': {'type': 'string', 'enum': ['strict', 'loose']},
                },
                'additionalProperties': False,
            },
        ]
    }

    def __init__(self):
        # remember and filter by these fields
        self.fields = ['imdb_id', 'tmdb_id', 'trakt_movie_id']
        self.keyword = 'seen_movies'

    # We run last to make sure we don't reject duplicates before all the other plugins get a chance to reject.
    @plugin.priority(plugin.PRIORITY_LAST)
    def on_task_filter(self, task, config):  # pylint: disable=W0221
        if not isinstance(config, dict):
            config = {'matching': config}
        # Reject all entries without
        if config.get('matching') == 'strict':
            for entry in task.entries:
                if not any(field in entry for field in self.fields):
                    logger.info(
                        'Rejecting {} because of missing movie (imdb, tmdb or trakt) id',
                        entry['title'],
                    )
                    entry.reject('missing movie (imdb, tmdb or trakt) id, strict')
        # call super
        super().on_task_filter(task, config.get('scope', True))
        # check that two copies of a movie have not been accepted this run
        accepted_ids = defaultdict(set)
        for entry in task.accepted:
            for field in self.fields:
                if field in entry:
                    if entry[field] in accepted_ids[field]:
                        entry.reject('already accepted %s %s once in task' % (field, entry[field]))
                        break
                    else:
                        accepted_ids[field].add(entry[field])

    def on_task_learn(self, task, config):
        if not isinstance(config, dict):
            config = {'matching': config}
        # call super
        super().on_task_learn(task, config.get('scope', True))


@event('plugin.register')
def register_plugin():
    plugin.register(FilterSeenMovies, 'seen_movies', api_ver=2)



import json

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='output.sns')

DEFAULT_TEMPLATE_VALUE = json.dumps(
    {
        'entry': {
            'title': '{{title}}',
            'url': '{{url}}',
            'original_url': '{{original_url}}',
            'series': '{{series_name}}',
            'series_id': '{{series_id}}',
        },
        'task': '{{task}}',
    }
)


class SNSNotification:
    """
    Emits SNS notifications of entries

    Optionally writes the torrent itself to S3

    Example configuration::

      sns:
        [aws_access_key_id: <AWS ACCESS KEY ID>] (will be taken from AWS_ACCESS_KEY_ID environment if not provided)
        [aws_secret_access_key: <AWS SECRET ACCESS KEY>] (will be taken from AWS_SECRET_ACCESS_KEY environment if
            not provided)
        [profile_name: <AWS PROFILE NAME>] (If provided, use this profile name instead of the default.)
        aws_region: <REGION>
        sns_topic_arn: <SNS ARN>
        [sns_notification_template: <TEMPLATE] (defaults to DEFAULT_TEMPLATE_VALUE)

    """

    schema = {
        'type': 'object',
        'properties': {
            'sns_topic_arn': {'type': 'string'},
            'sns_notification_template': {'type': 'string', 'default': DEFAULT_TEMPLATE_VALUE},
            'aws_access_key_id': {'type': 'string'},
            'aws_secret_access_key': {'type': 'string'},
            'aws_region': {'type': 'string'},
            'profile_name': {'type': 'string'},
        },
        'required': ['sns_topic_arn', 'aws_region'],
        'additionalProperties': False,
    }

    def on_task_start(self, task, config):
        # verify that we actually support Boto 3
        try:
            import boto3  # noqa
        except ImportError as e:
            logger.debug('Error importing boto3: {}', e)
            raise plugin.DependencyError(
                "sns", "boto3", "Boto3 module required. ImportError: %s" % e
            )

    # this has to run near the end of the plugin chain, because we
    # should notify after all other outputs.
    @plugin.priority(0)
    def on_task_output(self, task, config):
        sender = SNSNotificationEmitter(config)
        sender.send_notifications(task)


class SNSNotificationEmitter:
    def __init__(self, config):
        self.config = config
        import boto3

        self.boto3 = boto3

        self.sns_notification_template = self.config.get(
            'sns_notification_template', DEFAULT_TEMPLATE_VALUE
        )

    def build_session(self):
        self.session = self.boto3.Session(
            aws_access_key_id=self.config.get('aws_access_key_id', None),
            aws_secret_access_key=self.config.get('aws_secret_access_key', None),
            profile_name=self.config.get('profile_name', None),
            region_name=self.config['aws_region'],
        )

    def get_topic(self):
        self.build_session()
        sns = self.session.resource('sns')
        topic = sns.Topic(self.config['sns_topic_arn'])
        return topic

    def send_notifications(self, task):
        topic = self.get_topic()

        for entry in task.accepted:
            message = entry.render(self.sns_notification_template)
            if task.options.test:
                logger.info(
                    'SNS publication: region={}, arn={}', self.config['aws_region'], topic.arn
                )
                logger.info('Message: {}', message)
                continue

            try:
                response = topic.publish(Message=message)
            except Exception as e:
                logger.error('Error publishing {}: {}', entry['title'], e)
                continue
            else:
                logger.debug('Published {}: {}', entry, response)


@event('plugin.register')
def register_sns_plugin():
    plugin.register(SNSNotification, 'sns', api_ver=2)



from datetime import datetime, timedelta

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.manager import Session
from flexget.utils.tools import parse_timedelta

from . import db

SCHEMA_VER = 3
FAIL_LIMIT = 100

logger = logger.bind(name='failed')


class PluginFailed:
    """
    Records entry failures and stores them for trying again after a certain interval.
    Rejects them after they have failed too many times.

    """

    schema = {
        "oneOf": [
            # Allow retry_failed: no form to turn off plugin altogether
            {"type": "boolean"},
            {
                "type": "object",
                "properties": {
                    "retry_time": {"type": "string", "format": "interval", "default": "1 hour"},
                    "max_retries": {
                        "type": "integer",
                        "minimum": 0,
                        "maximum": FAIL_LIMIT,
                        "default": 3,
                    },
                    "retry_time_multiplier": {
                        # Allow turning off the retry multiplier with 'no' as well as 1
                        "oneOf": [{"type": "number", "minimum": 0}, {"type": "boolean"}],
                        "default": 1.5,
                    },
                },
                "additionalProperties": False,
            },
        ]
    }

    def prepare_config(self, config):
        if not isinstance(config, dict):
            config = {}
        config.setdefault('retry_time', '1 hour')
        config.setdefault('max_retries', 3)
        if config.get('retry_time_multiplier', True) is True:
            # If multiplier is not specified, or is specified as True, use the default
            config['retry_time_multiplier'] = 1.5
        else:
            # If multiplier is False, turn it off
            config['retry_time_multiplier'] = 1
        return config

    def retry_time(self, fail_count, config):
        """Return the timedelta an entry that has failed `fail_count` times before should wait before being retried."""
        base_retry_time = parse_timedelta(config['retry_time'])
        # Timedeltas do not allow floating point multiplication. Convert to seconds and then back to avoid this.
        base_retry_secs = base_retry_time.days * 86400 + base_retry_time.seconds
        retry_secs = base_retry_secs * (config['retry_time_multiplier'] ** fail_count)
        # prevent OverflowError: date value out of range, cap to 30 days
        max = 60 * 60 * 24 * 30
        if retry_secs > max:
            retry_secs = max
        return timedelta(seconds=retry_secs)

    @plugin.priority(plugin.PRIORITY_LAST)
    def on_task_input(self, task, config):
        if config is False:
            return
        config = self.prepare_config(config)
        for entry in task.all_entries:
            entry.on_fail(self.add_failed, config=config)

    def add_failed(self, entry, reason=None, config=None, **kwargs):
        """Adds entry to internal failed list, displayed with --failed"""
        # Make sure reason is a string, in case it is set to an exception instance
        reason = str(reason) or 'Unknown'
        with Session() as session:
            # query item's existence
            item = (
                session.query(db.FailedEntry)
                .filter(db.FailedEntry.title == entry['title'])
                .filter(db.FailedEntry.url == entry['original_url'])
                .first()
            )
            if not item:
                item = db.FailedEntry(entry['title'], entry['original_url'], reason)
                item.count = 0
            if item.count > FAIL_LIMIT:
                logger.error(
                    "entry with title '{}' has failed over {} times", entry['title'], FAIL_LIMIT
                )
                return
            retry_time = self.retry_time(item.count, config)
            item.retry_time = datetime.now() + retry_time
            item.count += 1
            item.tof = datetime.now()
            item.reason = reason
            session.merge(item)
            logger.debug('Marking {} in failed list. Has failed {} times.', item.title, item.count)
            if item.count <= config['max_retries']:
                plugin.get('backlog', self).add_backlog(
                    entry.task, entry, amount=retry_time, session=session
                )
            entry.task.rerun(plugin='retry_failed')

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_filter(self, task, config):
        if config is False:
            return
        config = self.prepare_config(config)
        max_count = config['max_retries']
        for entry in task.entries:
            item = (
                task.session.query(db.FailedEntry)
                .filter(db.FailedEntry.title == entry['title'])
                .filter(db.FailedEntry.url == entry['original_url'])
                .first()
            )
            if item:
                if item.count > max_count:
                    entry.reject(
                        'Has already failed %s times in the past. (failure reason: %s)'
                        % (item.count, item.reason)
                    )
                elif item.retry_time and item.retry_time > datetime.now():
                    entry.reject(
                        'Waiting before retrying entry which has failed in the past. (failure reason: %s)'
                        % item.reason
                    )


@event('plugin.register')
def register_plugin():
    plugin.register(PluginFailed, 'retry_failed', builtin=True, api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests
from flexget.utils.soup import get_soup

logger = logger.bind(name='pogcal')


class InputPogDesign:
    schema = {
        'type': 'object',
        'properties': {'username': {'type': 'string'}, 'password': {'type': 'string'}},
        'required': ['username', 'password'],
        'additionalProperties': False,
    }

    name_map = {
        'The Tonight Show [Leno]': 'The Tonight Show With Jay Leno',
        'Late Show [Letterman]': 'David Letterman',
    }

    def on_task_input(self, task, config):
        session = requests.Session()
        data = {
            'username': config['username'],
            'password': config['password'],
            'sub_login': 'Account Login',
        }
        try:
            r = session.post('https://www.pogdesign.co.uk/cat/login', data=data)
            if 'Login to Your Account' in r.text:
                raise plugin.PluginError('Invalid username/password for pogdesign.')
            page = session.get('https://www.pogdesign.co.uk/cat/show-select')
        except requests.RequestException as e:
            raise plugin.PluginError('Error retrieving source: %s' % e)
        soup = get_soup(page.text)
        entries = []
        for row in soup.find_all('li', {'class': 'selectgrp checked'}):

            # Get name
            t = row.find('strong')

            # Remove <span> tags
            spantags = t.find_all('span')
            for s in spantags:
                s.extract()
            t = row.text
            # remove newlines and whitepsace
            t = t.replace('\n', '').strip()

            if t.endswith('[The]'):
                t = 'The ' + t[:-6]

            # Make certain names friendlier
            if t in self.name_map:
                t = self.name_map[t]

            e = Entry()
            e['title'] = t
            e['url'] = 'https://www.pogdesign.co.uk/{0}'.format(
                row.find_next('a')['href'].lstrip('/')
            )
            entries.append(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(InputPogDesign, 'pogcal', api_ver=2)



import os

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.tools import aggregate_inputs

logger = logger.bind(name='torrent_match')


class TorrentMatchFile:
    def __init__(self, path, size):
        self.path = path
        self.size = size

    def __repr__(self):
        return "%s(path=%s, size=%s)" % (self.__class__.__name__, str(self.path), self.size)


class TorrentMatch:
    """Plugin that attempts to match .torrents to local files"""

    schema = {
        'type': 'object',
        'properties': {
            'what': {
                'type': 'array',
                'items': {
                    'allOf': [
                        {'$ref': '/schema/plugins?phase=input'},
                        {'maxProperties': 1, 'minProperties': 1},
                    ]
                },
            },
            'max_size_difference': {'type': 'string', 'format': 'percent', 'default': '0%'},
        },
        'required': ['what'],
        'additionalProperties': False,
    }

    def get_local_files(self, config, task):
        cwd = os.getcwd()  # save the current working directory
        entries = aggregate_inputs(task, config['what'])
        result = []
        for entry in entries:
            location = entry.get('location')
            if not location or not os.path.exists(location):
                logger.warning('{} is not a local file. Skipping.', entry['title'])
                entry.reject('not a local file')
                continue

            result.append(entry)
            entry['files'] = []

            if os.path.isfile(location):
                entry['files'].append(TorrentMatchFile(location, os.path.getsize(location)))
            else:
                # change working dir to make things simpler
                os.chdir(location)
                # traverse the file tree
                for root, _, files in os.walk('.'):
                    # we only need to iterate over files
                    for f in files:
                        file_path = os.path.join(root, f)
                        # We need normpath to strip out the dot
                        abs_file_path = os.path.normpath(os.path.join(location, file_path))
                        entry['files'].append(
                            TorrentMatchFile(abs_file_path, os.path.getsize(file_path))
                        )

        # restore the working directory
        os.chdir(cwd)

        return result

    # Run last in download phase to make sure we have downloaded .torrent to temp before modify phase
    @plugin.priority(0)
    def on_task_download(self, task, config):
        for entry in task.accepted:
            if 'file' not in entry and 'download' not in task.config:
                # If the download plugin is not enabled, we need to call it to get
                # our temp .torrent files
                plugin.get('download', self).get_temp_files(
                    task, handle_magnets=True, fail_html=True
                )

    def prepare_config(self, config):
        if not isinstance(config['max_size_difference'], float):
            config['max_size_difference'] = float(config['max_size_difference'].rstrip('%'))

        return config

    # Run after 'torrent' plugin, this is not really a modify plugin though, but we need 'torrent' field
    def on_task_modify(self, task, config):
        config = self.prepare_config(config)
        max_size_difference = config['max_size_difference']
        local_entries = self.get_local_files(config, task)

        matched_entries = set()
        for entry in task.accepted:
            if 'torrent' not in entry:
                logger.debug('Skipping entry {} as it is not a torrent file', entry['title'])
                continue

            # Find all files and file sizes in the .torrent.
            torrent_files = []
            for item in entry['torrent'].get_filelist():
                # if torrent is a multi_file, prepend the name
                path = os.path.join(item['path'], item['name'])
                if entry['torrent'].is_multi_file:
                    path = os.path.join(entry['torrent'].name, path)

                torrent_files.append(TorrentMatchFile(path, item['size']))

            # Iterate over the files/dirs from the  what  plugins
            for local_entry in local_entries:
                logger.debug(
                    'Checking local entry {} against {}', local_entry['title'], entry['title']
                )

                local_files = local_entry['files']

                # skip root dir of the local entry if torrent is single file
                has_root_dir = entry['torrent'].is_multi_file and entry['torrent'].name

                if not has_root_dir:  # single-file
                    torrent_file = torrent_files[0]
                    for local_file in local_files:
                        if (
                            torrent_file.path in str(local_file.path)
                            and torrent_file.size == local_file.size
                        ):
                            # if the filename with ext is contained in 'location', we must grab its parent as path
                            if os.path.basename(torrent_file.path) in str(local_entry['location']):
                                entry['path'] = os.path.dirname(local_entry['location'])
                            else:
                                entry['path'] = local_entry['location']
                            logger.debug('Path for {} set to {}', entry['title'], entry['path'])
                            matched_entries.add(entry)
                            break
                else:
                    matches = 0
                    missing_size = 0
                    total_size = 0
                    path = ''

                    candidate_files = []
                    # Find candidate files ie. files whose path contains the torrent name
                    for local_file in local_files:
                        if entry['torrent'].name in str(local_file.path):
                            # we need to find the path that contains the torrent name since it's multi-file
                            if not path:
                                # attempt to extract path from the absolute file path
                                path = local_file.path

                                while entry['torrent'].name in path:
                                    path = os.path.dirname(path)

                            candidate_files.append(local_file)

                    logger.debug('Path for {} will be set to {}', entry['title'], path)

                    for torrent_file in torrent_files:
                        for candidate in candidate_files:
                            if (
                                torrent_file.path in candidate.path
                                and torrent_file.size == candidate.size
                            ):
                                logger.debug(
                                    'Path {} matched local file path {}',
                                    torrent_file.path,
                                    candidate.path,
                                )
                                matches += 1
                                break
                        else:
                            logger.debug('No local paths matched {}', torrent_file.path)
                            missing_size += torrent_file.size
                        total_size += torrent_file.size

                    size_difference = missing_size / total_size * 100
                    # we allow torrents that either match entirely or if the total size difference is below a threshold
                    if matches == len(torrent_files) or max_size_difference >= size_difference:
                        matched_entries.add(entry)
                        # set the path of the torrent entry
                        entry['path'] = path

                        logger.debug('Torrent {} matched path {}', entry['title'], entry['path'])
                        # TODO keep searching for even better matches?
                        break

        for entry in set(task.accepted).difference(matched_entries):
            entry.reject(
                'No local files matched {}% of the torrent size'.format(100 - max_size_difference)
            )


@event('plugin.register')
def register_plugin():
    plugin.register(TorrentMatch, 'torrent_match', api_ver=2)



from flexget import plugin
from flexget.event import event

from . import seen as plugin_seen


class FilterSeenInfoHash(plugin_seen.FilterSeen):
    """Prevents the same torrent from being downloaded twice by remembering the infohash of all downloaded torrents."""

    schema = {'oneOf': [{'type': 'boolean'}, {'type': 'string', 'enum': ['global', 'local']}]}

    def __init__(self):
        # remember and filter by these fields
        self.fields = ['torrent_info_hash']
        self.keyword = 'seen_info_hash'

    @plugin.priority(180)
    def on_task_filter(self, task, config):  # pylint: disable=W0221
        # Return if we are disabled.
        if config is False:
            return
        # First make sure all the torrent_info_hash fields are in upper case
        for entry in task.entries:
            if isinstance(entry.get('torrent_info_hash'), str):
                entry['torrent_info_hash'] = entry['torrent_info_hash'].upper()
        plugin_seen.FilterSeen.on_task_filter(self, task, config, remember_rejected=True)

    def on_task_modify(self, task, config):
        # Return if we are disabled.
        if config is False:
            return
        # Run the filter again after the torrent plugin has populated the infohash
        self.on_task_filter(task, config)
        # Make sure no duplicates were accepted this run
        accepted_infohashes = set()
        for entry in task.accepted:
            if 'torrent_info_hash' in entry:
                infohash = entry['torrent_info_hash']
                if infohash in accepted_infohashes:
                    entry.reject('Already accepted torrent with this infohash once for this task')
                else:
                    accepted_infohashes.add(infohash)


@event('plugin.register')
def register_plugin():
    plugin.register(FilterSeenInfoHash, 'seen_info_hash', builtin=True, api_ver=2)



import json


class TestValidator:
    config = '{tasks: {}}'

    def test_invalid_payload(self, api_client):
        new_task = {
            'name': 'new_task',
            'config': {
                'invalid_plugin': [{'title': 'entry 1'}],
                'fake_plugin2': {'url': 'http://test/rss'},
            },
        }

        rsp = api_client.json_post('/tasks/', data=json.dumps(new_task))

        assert rsp.status_code == 422
        data = json.loads(rsp.get_data(as_text=True))
        assert data.get('status_code') == 422
        assert data.get('message') == 'validation error'
        assert data.get('validation_errors')
        assert 'The keys' in data['validation_errors'][0]['message']
        assert 'invalid_plugin' in data['validation_errors'][0]['message']
        assert 'fake_plugin2' in data['validation_errors'][0]['message']



from collections.abc import MutableMapping
from typing import Any, Callable, Iterable, List, Mapping, NamedTuple, Sequence

from loguru import logger

logger = logger.bind(name='lazy_lookup')


class LazyCallee(NamedTuple):
    func: Callable
    keys: Sequence
    args: Sequence
    kwargs: Mapping


class LazyLookup:
    """
    This class stores the information to do a lazy lookup for a LazyDict. An instance is stored as a placeholder value
    for any key that can be lazily looked up. There should be one instance of this class per LazyDict.
    """

    def __init__(self, store: 'LazyDict') -> None:
        self.store = store
        self.callee_list: List[LazyCallee] = []

    def add_func(self, func: Callable, keys: Sequence, args: Sequence, kwargs: Mapping) -> None:
        self.callee_list.append(LazyCallee(func, keys, args, kwargs))

    def __getitem__(self, key) -> Any:
        from flexget.plugin import PluginError

        while self.store.is_lazy(key):
            index = next(
                (i for i, callee in enumerate(self.callee_list) if key in callee.keys), None
            )
            if index is None:
                # All lazy lookup functions for this key were tried unsuccessfully
                return None
            callee = self.callee_list.pop(index)
            try:
                callee.func(self.store, *(callee.args or []), **(callee.kwargs or {}))
            except PluginError as e:
                e.logger.info(e)
            except Exception as e:
                logger.error('Unhandled error in lazy lookup plugin: {}', e)
                from flexget.manager import manager

                if manager:
                    manager.crash_report()
                else:
                    logger.opt(exception=True).debug('Traceback')
        return self.store[key]

    def __repr__(self):
        return '<LazyLookup(%r)>' % self.callee_list


class LazyDict(MutableMapping):
    def __init__(self, *args, **kwargs):
        self.store = dict(*args, **kwargs)

    def __setitem__(self, key, value):
        self.store[key] = value

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store)

    def __delitem__(self, key):
        del self.store[key]

    def __getitem__(self, key):
        item = self.store[key]
        if isinstance(item, LazyLookup):
            return item[key]
        return item

    def __copy__(self):
        return type(self)(self.store)

    copy = __copy__

    def get(
        self, key, default: Any = None, eval_lazy: bool = True
    ) -> Any:  # pylint: disable=W0221
        """
        Adds the `eval_lazy` keyword argument to the normal :func:`dict.get` method.

        :param bool eval_lazy: If False, the default will be returned rather than evaluating a lazy field.
        """
        item = self.store.get(key, default)
        if isinstance(item, LazyLookup):
            if eval_lazy:
                try:
                    return item[key]
                except KeyError:
                    return default
            else:
                return default
        return item

    @property
    def _lazy_lookup(self) -> LazyLookup:
        """
        The LazyLookup instance for this LazyDict.
        If one is already stored in this LazyDict, it is returned, otherwise a new one is instantiated.
        """
        for val in self.store.values():
            if isinstance(val, LazyLookup):
                return val
        return LazyLookup(self)

    def register_lazy_func(
        self, func: Callable[[Mapping], None], keys: Iterable, args: Sequence, kwargs: Mapping
    ):
        """Register a list of fields to be lazily loaded by callback func.

        :param func:
          Callback function which is called when lazy key needs to be evaluated.
          Function call will get this LazyDict instance as a parameter.
          See :class:`LazyLookup` class for more details.
        :param keys:
          List of key names that `func` can provide.
        :param args: Arguments which will be passed to `func` when called.
        :param kwargs: Keyword arguments which will be passed to `func` when called.
        """
        ll = self._lazy_lookup
        ll.add_func(func, keys, args, kwargs)
        for key in keys:
            if key not in self.store:
                self[key] = ll

    def is_lazy(self, key) -> bool:
        """
        :param key: Key to check
        :return: True if value for key is lazy loading.
        :rtype: bool
        """
        return isinstance(self.store.get(key), LazyLookup)



"""Plugin for telegram input"""
import re

from loguru import logger
from requests.exceptions import HTTPError, RequestException

from flexget.config_schema import one_or_more
from flexget import plugin
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='from_telegram')

_TELEGRAM_API_URL = "https://api.telegram.org/"


class TelegramInput:
    """
    Parse any messages from Telegram and fills fields with regex

    Example:
      token: <token>
      only_new: yes
      whitelist:
        - username: <my_username>
        - group: <my_group>
        - fullname:
            first: <my_first_name>
            sur: <my_surname>
      entry:
        <field>: <regexp to match value>

    Note: If not declared, title will be the message
    """

    schema = {
        'type': 'object',
        'properties': {
            'token': {'type': 'string'},
            'types': one_or_more({'type': 'string', 'enum': ['private', 'group']}),
            'whitelist': {
                'type': 'array',
                'minItems': 1,
                'items': {
                    'oneOf': [
                        {
                            'type': 'object',
                            'properties': {'username': {'type': 'string'}},
                            'required': ['username'],
                            'additionalProperties': False,
                        },
                        {
                            'type': 'object',
                            'properties': {
                                'fullname': {
                                    'type': 'object',
                                    'properties': {
                                        'first': {'type': 'string'},
                                        'sur': {'type': 'string'},
                                    },
                                    'required': ['first', 'sur'],
                                    'additionalProperties': False,
                                }
                            },
                            'required': ['fullname'],
                            'additionalProperties': False,
                        },
                        {
                            'type': 'object',
                            'properties': {'group': {'type': 'string'}},
                            'required': ['group'],
                            'additionalProperties': False,
                        },
                    ]
                },
            },
            'only_new': {'type': 'boolean', 'default': True},
            'entry': {
                'type': 'object',
                'properties': {
                    'url': {'type': 'string', 'format': 'regex'},
                    'title': {'type': 'string', 'format': 'regex'},
                },
                'additionalProperties': {'type': 'string', 'format': 'regex'},
            },
        },
        'required': ['token'],
        'additonalProperties': False,
    }

    @plugin.internet(logger)
    def on_task_input(self, task, config):
        # Load The Configs
        token = config['token']
        only_new = config['only_new']
        entry_config = config.get('entry')
        whitelist = config.get('whitelist', [])
        types = config.get('types', ['private', 'group'])

        # Get Last Checked ID
        persistence_name = f"{token}_update_id"
        update_id = task.simple_persistence.get(persistence_name)

        # Get only new messages
        params = {}
        if update_id and only_new:
            update_id += 1
            params['offset'] = update_id

        # The Target URL
        url = f"{_TELEGRAM_API_URL}bot{token}/getUpdates"

        # Get Telegram Updates
        try:
            response = task.requests.get(url, timeout=60, raise_status=True, params=params).json()
        except HTTPError as e:
            raise plugin.PluginError(f"Error getting telegram update: {e}")

        # We have a error
        if not response['ok']:
            raise plugin.PluginError(
                f"Telegram updater returned error {response['error_code']}: {response['description']}"
            )

        # Get All New Messages
        messages = response['result']

        entries = []
        for message in messages:
            # This is the ID
            update_id = message['update_id']

            # Update the last ID for the Bot
            logger.debug("Last Update set to {}", update_id)
            task.simple_persistence[persistence_name] = update_id

            # We Don't care if it's not a message or no text
            if (
                'message' not in message
                or 'text' not in message['message']
                or 'chat' not in message['message']
                or 'type' not in message['message']['chat']
            ):
                logger.debug("Invalid message discarted: {}", message)
                continue

            logger.debug("Income message: {}", message)

            # Check Types
            if types and message['message']['chat']['type'] not in types:
                logger.debug("Ignoring message because of invalid type {}", message)
                continue

            # Create Base Entry
            text = message['message']['text']
            entry = Entry()
            entry['title'] = text

            # We need a url, so we add a dummy
            entry['url'] = f"http://localhost?update_id={str(update_id)}"

            # Store the message if we need to use it in other plugins
            entry['telegram_message'] = message['message']

            # Check From
            message_from = message['message']['from']
            message_chat = message['message']['chat']

            if whitelist:
                for check in whitelist:
                    if 'username' in check and check['username'] == message_from['username']:
                        logger.debug("WhiteListing: Username {}", message_from['username'])
                        break
                    elif (
                        'fullname' in check
                        and check['fullname']['first'] == message_from['first_name']
                        and check['fullname']['sur'] == message_from['last_name']
                    ):
                        logger.debug(
                            "WhiteListing: Full Name {} {}",
                            message_from['first_name'],
                            message_from['last_name'],
                        )
                        break
                    elif 'group' in check:
                        if (
                            message_chat['type'] == 'group'
                            and message_chat['title'] == check['group']
                        ):
                            logger.debug("WhiteListing: Group {}", message_chat['title'])
                            break
                else:
                    logger.debug("Ignoring message because of no whitelist match {}", message)
                    continue

            # Process the entry config
            accept = True
            if entry_config:
                for field, regexp in entry_config.items():
                    match = re.search(regexp, text)
                    if match:
                        try:
                            # Add field to entry
                            entry[field] = match.group(1)
                        except IndexError:
                            logger.error(
                                'Regex for field `{}` must contain a capture group', field
                            )
                            raise plugin.PluginError(
                                'Your from_telegram plugin config contains errors, please correct them.'
                            )
                    else:
                        logger.debug('Ignored entry, not match on field {}: {}', field, entry)
                        accept = False
                        break

            # Append the entry
            if accept:
                entries.append(entry)
                logger.debug('Added entry {}', entry)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(TelegramInput, 'from_telegram', api_ver=2)



import datetime
import re

from loguru import logger
from requests.exceptions import TooManyRedirects
from sqlalchemy import Column, DateTime, Unicode

from flexget import db_schema, plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session
from flexget.utils.database import json_synonym
from flexget.utils.requests import RequestException
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='morethantv')
Base = db_schema.versioned_base('morethantv', 0)

requests = RequestSession()
requests.add_domain_limiter(
    TimedLimiter('morethan.tv', '5 seconds')
)  # TODO find out if they want a delay

CATEGORIES = {'Movies': 'filter_cat[1]', 'TV': 'filter_cat[2]', 'Other': 'filter_cat[3]'}

TAGS = [
    'action',
    'adventure',
    'animation',
    'anime',
    'art',
    'asian',
    'biography',
    'celebrities',
    'comedy',
    'cooking',
    'crime',
    'cult',
    'documentary',
    'drama',
    'educational',
    'elclasico',
    'family',
    'fantasy',
    'film.noir',
    'filmromanesc',
    'food',
    'football',
    'formula.e',
    'formula1',
    'gameshow',
    'highlights',
    'history',
    'horror',
    'investigation',
    'lifestyle',
    'liga1',
    'ligabbva',
    'ligue1',
    'martial.arts',
    'morethan.tv',
    'motogp',
    'musical',
    'mystery',
    'nba',
    'news',
    'other',
    'performance',
    'philosophy',
    'politics',
    'reality',
    'romance',
    'romanian.content',
    'science',
    'scifi',
    'short',
    'silent',
    'sitcom',
    'sketch',
    'sports',
    'talent',
    'tennis',
    'thriller',
    'uefachampionsleague',
    'uefaeuropaleague',
    'ufc',
    'war',
    'western',
    'wta',
]


class MoreThanTVCookie(Base):
    __tablename__ = 'morethantv_cookie'

    username = Column(Unicode, primary_key=True)
    _cookie = Column('cookie', Unicode)
    cookie = json_synonym('_cookie')
    expires = Column(DateTime)


class SearchMoreThanTV:
    """
    MorethanTV search plugin.
    """

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'category': one_or_more(
                {'type': 'string', 'enum': list(CATEGORIES.keys())}, unique_items=True
            ),
            'order_by': {
                'type': 'string',
                'enum': ['seeders', 'leechers', 'time'],
                'default': 'time',
            },
            'order_way': {'type': 'string', 'enum': ['desc', 'asc'], 'default': 'desc'},
            'tags': one_or_more({'type': 'string', 'enum': TAGS}, unique_items=True),
            'all_tags': {'type': 'boolean', 'default': True},
        },
        'required': ['username', 'password'],
        'additionalProperties': False,
    }

    base_url = 'https://www.morethan.tv/'
    errors = False

    def get(self, url, params, username, password, force=False):
        """
        Wrapper to allow refreshing the cookie if it is invalid for some reason

        :param str url:
        :param list params:
        :param str username:
        :param str password:
        :param bool force: flag used to refresh the cookie forcefully ie. forgo DB lookup
        :return:
        """
        cookies = self.get_login_cookie(username, password, force=force)
        invalid_cookie = False

        try:
            response = requests.get(url, params=params, cookies=cookies)
            if self.base_url + 'login.php' in response.url:
                invalid_cookie = True
        except TooManyRedirects:
            # Apparently it endlessly redirects if the cookie is invalid?
            logger.debug('MoreThanTV request failed: Too many redirects. Invalid cookie?')
            invalid_cookie = True

        if invalid_cookie:
            if self.errors:
                raise plugin.PluginError(
                    'MoreThanTV login cookie is invalid. Login page received?'
                )
            self.errors = True
            # try again
            response = self.get(url, params, username, password, force=True)
        else:
            self.errors = False

        return response

    def get_login_cookie(self, username, password, force=False):
        """
        Retrieves login cookie

        :param str username:
        :param str password:
        :param bool force: if True, then retrieve a fresh cookie instead of looking in the DB
        :return:
        """
        if not force:
            with Session() as session:
                saved_cookie = (
                    session.query(MoreThanTVCookie)
                    .filter(MoreThanTVCookie.username == username)
                    .first()
                )
                if (
                    saved_cookie
                    and saved_cookie.expires
                    and saved_cookie.expires >= datetime.datetime.now()
                ):
                    logger.debug('Found valid login cookie')
                    return saved_cookie.cookie

        url = self.base_url + 'login.php'
        try:
            logger.debug('Attempting to retrieve MoreThanTV cookie')
            response = requests.post(
                url,
                data={
                    'username': username,
                    'password': password,
                    'login': 'Log in',
                    'keeplogged': '1',
                },
                timeout=30,
            )
        except RequestException as e:
            raise plugin.PluginError('MoreThanTV login failed: %s' % e)

        if 'Your username or password was incorrect.' in response.text:
            raise plugin.PluginError(
                'MoreThanTV login failed: Your username or password was incorrect.'
            )

        with Session() as session:
            expires = None
            for c in requests.cookies:
                if c.name == 'session':
                    expires = c.expires
            if expires:
                expires = datetime.datetime.fromtimestamp(expires)
            logger.debug('Saving or updating MoreThanTV cookie in db')
            cookie = MoreThanTVCookie(
                username=username, cookie=dict(requests.cookies), expires=expires
            )
            session.merge(cookie)
            return cookie.cookie

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for entries on MoreThanTV
        """
        params = {}

        if 'category' in config:
            categories = (
                config['category']
                if isinstance(config['category'], list)
                else [config['category']]
            )
            for category in categories:
                params[CATEGORIES[category]] = 1

        if 'tags' in config:
            tags = config['tags'] if isinstance(config['tags'], list) else [config['tags']]
            tags = ', '.join(tags)
            params['taglist'] = tags

        entries = set()

        params.update(
            {
                'tags_type': int(config['all_tags']),
                'order_by': config['order_by'],
                'search_submit': 1,
                'order_way': config['order_way'],
                'action': 'basic',
                'group_results': 0,
            }
        )

        for search_string in entry.get('search_strings', [entry['title']]):
            params['searchstr'] = search_string.replace("'", "")
            logger.debug('Using search params: {}', params)
            try:
                page = self.get(
                    self.base_url + 'torrents.php', params, config['username'], config['password']
                )
                logger.debug('requesting: {}', page.url)
            except RequestException as e:
                logger.error('MoreThanTV request failed: {}', e)
                continue

            soup = get_soup(page.content)
            for result in soup.findAll('tr', attrs={'class': 'torrent'}):
                group_info = result.find('td', attrs={'class': 'big_info'}).find(
                    'div', attrs={'class': 'group_info'}
                )
                # TODO: compiling regexp within loop achieves nothing
                title = group_info.find('a', href=re.compile(r'torrents.php\?id=\d+')).text
                url = (
                    self.base_url
                    + group_info.find('a', href=re.compile(r'torrents.php\?action=download'))[
                        'href'
                    ]
                )
                torrent_info = result.findAll('td', attrs={'class': 'number_column'})
                size = re.search(r'(\d+(?:[.,]\d+)*)\s?([KMG]B)', torrent_info[0].text)
                torrent_tags = ', '.join(
                    [tag.text for tag in group_info.findAll('div', attrs={'class': 'tags'})]
                )

                e = Entry()

                e['title'] = title
                e['url'] = url
                e['torrent_snatches'] = int(torrent_info[1].text)
                e['torrent_seeds'] = int(torrent_info[2].text)
                e['torrent_leeches'] = int(torrent_info[3].text)
                e['torrent_internal'] = (
                    True if group_info.find('span', attrs={'class': 'flag_internal'}) else False
                )
                e['torrent_fast_server'] = (
                    True if group_info.find('span', attrs={'class': 'flag_fast'}) else False
                )
                e['torrent_sticky'] = (
                    True if group_info.find('span', attrs={'class': 'flag_sticky'}) else False
                )
                e['torrent_tags'] = torrent_tags

                e['content_size'] = parse_filesize(size.group(0))

                entries.add(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchMoreThanTV, 'morethantv', interfaces=['search'], api_ver=2)



import re

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils.tools import ReList

logger = logger.bind(name='regex_extract')


class RegexExtract:
    r"""
    Updates an entry with the values of regex matched named groups

    Usage:

      regex_extract:
        field: <string>
        regex:
          - <regex>
        [prefix]: <string>


    Example:

      regex_extract:
        prefix: f1_
        field: title
        regex:
          - Formula\.?1.(?P<location>*?)

    """

    schema = {
        'type': 'object',
        'properties': {
            'prefix': {'type': 'string'},
            'field': {'type': 'string'},
            'regex': one_or_more({'type': 'string', 'format': 'regex'}),
        },
    }

    def on_task_start(self, task, config):
        regex = config.get('regex')
        if isinstance(regex, str):
            regex = [regex]
        self.regex_list = ReList(regex)

        # Check the regex
        try:
            for _ in self.regex_list:
                pass
        except re.error as e:
            raise plugin.PluginError('Error compiling regex: %s' % str(e))

    def on_task_modify(self, task, config):

        prefix = config.get('prefix')
        modified = 0

        for entry in task.entries:
            for rx in self.regex_list:
                entry_field = entry.get('title')
                logger.debug('Matching {} with regex: {}', entry_field, rx)
                try:
                    match = rx.match(entry_field)
                except re.error as e:
                    raise plugin.PluginError('Error encountered processing regex: %s' % str(e))
                if match:
                    logger.debug('Successfully matched {}', entry_field)
                    data = match.groupdict()
                    if prefix:
                        for key in list(data.keys()):
                            data[prefix + key] = data[key]
                            del data[key]
                    logger.debug('Values added to entry: {}', data)
                    entry.update(data)
                    modified += 1

        logger.info('{} entries matched and modified', modified)


@event('plugin.register')
def register_plugin():
    plugin.register(RegexExtract, 'regex_extract', api_ver=2)



from flexget.api.app import base_message
from flexget.utils import json


class TestFormatChecker:
    config = 'tasks: {}'

    def test_quality(self, api_client, schema_match):
        payload1 = {'quality': '720p'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'quality': '720p-1080p'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_quality_req(self, api_client, schema_match):
        payload1 = {'quality_requirements': '720p-1080p'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'quality': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_time(self, api_client, schema_match):
        payload = {'time': '10:00'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload = {'time': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_interval(self, api_client, schema_match):
        payload1 = {'interval': '1 day'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'interval': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_percent(self, api_client, schema_match):
        payload1 = {'percent': '79%'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'percent': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_size(self, api_client, schema_match):
        payload1 = {'size': '4GB'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'percent': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_regex(self, api_client, schema_match):
        payload1 = {'regex': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'regex': '(('}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_file(self, api_client, schema_match):
        payload1 = {'file': 'test_format_checker_api.py'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'file': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_path(self, api_client, schema_match):
        payload1 = {'path': '../api_tests'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'path': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_url(self, api_client, schema_match):
        payload1 = {'url': 'http://google.com'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'url': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_episode_identifier(self, api_client, schema_match):
        payload1 = {'episode_identifier': 's01e01'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'episode_identifier': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_episode_or_season_id(self, api_client, schema_match):
        payload1 = {'episode_or_season_id': 's01'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload1))
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        payload2 = {'episode_or_season_id': 'bla'}

        rsp = api_client.json_post('/format_check/', data=json.dumps(payload2))
        assert rsp.status_code == 422, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors



# make importing these a bit less hassle
from flexget.utils.parsers.movie import MovieParser  # noqa pylint: disable=unused-import
from flexget.utils.parsers.parser import TitleParser  # noqa pylint: disable=unused-import
from flexget.utils.parsers.series import (  # noqa pylint: disable=unused-import
    ID_TYPES,
    SeriesParser,
)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='torrent_size')


class TorrentSize:
    """
    Provides file size information when dealing with torrents
    """

    @plugin.priority(200)
    def on_task_modify(self, task, config):
        for entry in task.entries:
            if 'torrent' in entry:
                size = entry['torrent'].size / 1024 / 1024
                logger.debug('{} size: {} MB', entry['title'], size)
                entry['content_size'] = size


@event('plugin.register')
def register_plugin():
    plugin.register(TorrentSize, 'torrent_size', builtin=True, api_ver=2)



from datetime import datetime, timedelta

from dateutil import parser
from loguru import logger
from requests.exceptions import RequestException
from sqlalchemy import (
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
    Table,
    Unicode,
    and_,
    or_,
)
from sqlalchemy.orm import relation
from sqlalchemy.orm.exc import MultipleResultsFound

from flexget import db_schema, plugin
from flexget.event import event
from flexget.utils import requests
from flexget.utils.database import json_synonym, with_session
from flexget.utils.tools import split_title_year

logger = logger.bind(name='api_tvmaze')

DB_VERSION = 7
Base = db_schema.versioned_base('tvmaze', DB_VERSION)
UPDATE_INTERVAL = 7  # Used for expiration, number is in days
BASE_URL = 'https://api.tvmaze.com'

TVMAZE_SHOW_PATH = "/shows/{}"
TVMAZE_LOOKUP_PATH = "/lookup/shows"
TVMAZE_SEARCH_PATH = "/singlesearch/shows"
TVMAZE_EPISODES_BY_DATE_PATH = "/shows/{}/episodesbydate"
TVMAZE_EPISODES_BY_NUMBER_PATH = "/shows/{}/episodebynumber"
TVMAZE_SEASONS = '/shows/{}/seasons'


@db_schema.upgrade('tvmaze')
def upgrade(ver, session):
    if ver is None or ver < 7:
        raise db_schema.UpgradeImpossible
    return ver


class TVMazeGenre(Base):
    __tablename__ = 'tvmaze_genres'

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(Unicode, unique=True)


genres_table = Table(
    'tvmaze_series_genres',
    Base.metadata,
    Column('series_id', Integer, ForeignKey('tvmaze_series.tvmaze_id')),
    Column('genre_id', Integer, ForeignKey('tvmaze_genres.id')),
)

Base.register_table(genres_table)


class TVMazeLookup(Base):
    __tablename__ = 'tvmaze_lookup'

    id = Column(Integer, primary_key=True, autoincrement=True)
    search_name = Column(Unicode, index=True, unique=True)
    series_id = Column(Integer, ForeignKey('tvmaze_series.tvmaze_id'))
    series = relation('TVMazeSeries', backref='search_strings')

    def __init__(self, search_name, series_id=None, series=None):
        self.search_name = search_name.lower()
        if series_id:
            self.series_id = series_id
        if series:
            self.series = series

    def __repr__(self):
        return '<TVMazeLookup(search_name={0},series_id={1})'.format(
            self.search_name, self.series_id
        )


class TVMazeSeries(Base):
    __tablename__ = 'tvmaze_series'

    tvmaze_id = Column(Integer, primary_key=True)
    status = Column(Unicode)
    rating = Column(Float)
    genres = relation(TVMazeGenre, secondary=genres_table)
    weight = Column(Integer)
    updated = Column(DateTime)  # last time show was updated at tvmaze
    name = Column(Unicode)
    language = Column(Unicode)
    _schedule = Column('schedule', Unicode)
    schedule = json_synonym('_schedule')
    url = Column(String)
    original_image = Column(String)
    medium_image = Column(String)
    tvdb_id = Column(Integer)
    tvrage_id = Column(Integer)
    premiered = Column(DateTime)
    year = Column(Integer)
    summary = Column(Unicode)
    webchannel = Column(String)
    runtime = Column(Integer)
    show_type = Column(String)
    network = Column(Unicode)
    episodes = relation(
        'TVMazeEpisodes',
        order_by='TVMazeEpisodes.season_number',
        cascade='all, delete, delete-orphan',
        backref='series',
    )
    seasons = relation(
        'TVMazeSeason',
        order_by='TVMazeSeason.number',
        cascade='all, delete, delete-orphan',
        backref='series',
    )

    last_update = Column(DateTime)  # last time we updated the db for the show

    def __init__(self, series, session):
        self.tvmaze_id = series['id']
        self.update(series, session)

    def to_dict(self):
        return {
            'tvmaze_id': self.tvmaze_id,
            'status': self.status,
            'rating': self.rating,
            'genres': [genre.name for genre in self.genres],
            'weight': self.weight,
            'updated': self.updated,
            'name': self.name,
            'language': self.language,
            'schedule': self.schedule,
            'url': self.url,
            'original_image': self.original_image,
            'medium_image': self.medium_image,
            'tvdb_id': self.tvdb_id,
            'tvrage_id': self.tvrage_id,
            'premiered': self.premiered,
            'year': self.year,
            'summary': self.summary,
            'webchannel': self.webchannel,
            'runtime': self.runtime,
            'show_type': self.show_type,
            'network': self.network,
            'last_update': self.last_update,
        }

    def update(self, series, session):
        self.status = series['status']
        self.rating = series['rating']['average']
        self.weight = series['weight']
        self.updated = datetime.fromtimestamp(series['updated'])
        self.name = series['name']
        self.language = series['language']
        self.schedule = series['schedule']
        self.url = series['url']
        self.original_image = series.get('image').get('original') if series.get('image') else None
        self.medium_image = series.get('image').get('medium') if series.get('image') else None
        self.tvdb_id = series['externals'].get('thetvdb')
        self.tvrage_id = series['externals'].get('tvrage')
        self.premiered = (
            parser.parse(series.get('premiered'), ignoretz=True)
            if series.get('premiered')
            else None
        )
        self.year = int(series.get('premiered')[:4]) if series.get('premiered') else None
        self.summary = series['summary']
        self.webchannel = series.get('web_channel')['name'] if series.get('web_channel') else None
        self.runtime = series['runtime']
        self.show_type = series['type']
        self.network = series.get('network')['name'] if series.get('network') else None
        self.last_update = datetime.now()

        self.genres = get_db_genres(series['genres'], session)
        self.seasons = self.populate_seasons(series)

    def __repr__(self):
        return '<TVMazeSeries(title=%s,id=%s,last_update=%s)>' % (
            self.name,
            self.tvmaze_id,
            self.last_update,
        )

    def __str__(self):
        return self.name

    @property
    def expired(self):
        if not self.last_update:
            logger.debug('no last update attribute, series set for update')
            return True
        time_dif = datetime.now() - self.last_update
        expiration = time_dif.days > UPDATE_INTERVAL
        return expiration

    def populate_seasons(self, series=None):
        if series and '_embedded' in series and series['_embedded'].get('seasons'):
            seasons = series['_embedded']['seasons']
        else:
            seasons = get_seasons(self.tvmaze_id)
        return [TVMazeSeason(season, self.tvmaze_id) for season in seasons]


class TVMazeSeason(Base):
    __tablename__ = 'tvmaze_seasons'

    tvmaze_id = Column(Integer, primary_key=True)
    series_id = Column(Integer, ForeignKey('tvmaze_series.tvmaze_id'), nullable=False)

    number = Column(Integer)
    url = Column(String)
    name = Column(Unicode)
    episode_order = Column(Integer)
    airdate = Column(DateTime)
    end_date = Column(DateTime)
    network = Column(Unicode)
    web_channel = Column(Unicode)
    image = Column(String)
    summary = Column(Unicode)

    def __init__(self, season, series_id):
        self.tvmaze_id = season['id']
        self.series_id = series_id
        self.number = season['number']
        self.update(season)

    def update(self, season):
        self.url = season['url']
        self.name = season['name']
        self.end_date = (
            parser.parse(season.get('endDate'), ignoretz=True) if season.get('endDate') else None
        )
        self.airdate = (
            parser.parse(season['premiereDate'], ignoretz=True)
            if season.get('premiereDate')
            else None
        )
        self.web_channel = season['web_channel']['name'] if season.get('web_channel') else None
        self.network = season['network']['name'] if season.get('network') else None
        self.image = season['image']['original'] if season.get('image') else None
        self.summary = season['summary']


class TVMazeEpisodes(Base):
    __tablename__ = 'tvmaze_episode'

    tvmaze_id = Column(Integer, primary_key=True)
    series_id = Column(Integer, ForeignKey('tvmaze_series.tvmaze_id'), nullable=False)
    number = Column(Integer, nullable=False)
    season_number = Column(Integer, nullable=False)

    title = Column(Unicode)
    airdate = Column(DateTime)
    url = Column(String)
    original_image = Column(String)
    medium_image = Column(String)
    airstamp = Column(DateTime)
    runtime = Column(Integer)
    summary = Column(Unicode)
    last_update = Column(DateTime)

    def to_dict(self):
        return {
            'tvmaze_id': self.tvmaze_id,
            'series_id': self.series_id,
            'number': self.number,
            'season_number': self.season_number,
            'title': self.title,
            'airdate': self.airdate,
            'url': self.url,
            'original_image': self.original_image,
            'medium_image': self.medium_image,
            'airstamp': self.airstamp,
            'runtime': self.runtime,
            'summary': self.summary,
            'last_update': self.last_update,
        }

    def __init__(self, episode, series_id):
        self.series_id = series_id
        self.tvmaze_id = episode['id']
        self.season_number = episode['season']
        self.number = episode['number']
        self.update(episode)

    def update(self, episode):
        self.summary = episode['summary']
        self.title = episode['name']
        self.airdate = (
            datetime.strptime(episode.get('airdate'), '%Y-%m-%d')
            if episode.get('airdate')
            else None
        )
        self.url = episode['url']
        self.original_image = (
            episode.get('image').get('original') if episode.get('image') else None
        )
        self.medium_image = episode.get('image').get('medium') if episode.get('image') else None
        self.airstamp = (
            parser.parse(episode.get('airstamp'), ignoretz=True)
            if episode.get('airstamp')
            else None
        )
        self.runtime = episode['runtime']
        self.last_update = datetime.now()

    @property
    def expired(self):
        if not self.last_update:
            logger.debug('no last update attribute, episode set for update')
            return True
        time_dif = datetime.now() - self.last_update
        expiration = time_dif.days > UPDATE_INTERVAL
        if expiration:
            logger.debug(
                'episode {}, season {} for series {} is expired.',
                self.number,
                self.season_number,
                self.series_id,
            )
        return expiration


def get_db_genres(genres, session):
    db_genres = []
    for genre in genres:
        db_genre = session.query(TVMazeGenre).filter(TVMazeGenre.name == genre).first()
        if not db_genre:
            db_genre = TVMazeGenre(name=genre)
            logger.trace('adding genre {} to db', genre)
            session.add(db_genre)
        else:
            logger.trace('genre {} found in db, returning', db_genre.name)
        db_genres.append(db_genre)
    return db_genres


def search_params_for_series(**lookup_params):
    search_params = {
        'tvmaze_id': lookup_params.get('tvmaze_id'),
        'tvdb_id': lookup_params.get('tvdb_id'),
        'tvrage_id': lookup_params.get('tvrage_id'),
        'name': lookup_params.get('title') or lookup_params.get('series_name'),
    }
    logger.debug('returning search params for series lookup: {}', search_params)
    return search_params


@with_session
def from_cache(session=None, search_params=None, cache_type=None):
    """
    Returns a result from requested table based on search params

    :param session: Current session
    :param search_params: Relevant search params. Should match table column names
    :param cache_type: Object for search
    :return: Query result
    """
    if not any(search_params.values()):
        raise LookupError('No parameters sent for cache lookup')
    else:
        logger.debug(
            'searching db {} for the values {}',
            cache_type.__tablename__,
            list(search_params.items()),
        )
        result = (
            session.query(cache_type)
            .filter(
                or_(getattr(cache_type, col) == val for col, val in search_params.items() if val)
            )
            .first()
        )
    return result


@with_session
def from_lookup(session=None, title=None):
    logger.debug('searching lookup table using title {}', title)
    return session.query(TVMazeLookup).filter(TVMazeLookup.search_name == title.lower()).first()


@with_session
def add_to_lookup(session=None, title=None, series=None):
    logger.debug('trying to add search title {} to series {} in lookup table', title, series.name)
    exist = session.query(TVMazeLookup).filter(TVMazeLookup.search_name == title.lower()).first()
    if exist:
        logger.debug(
            'title {} already exist for series {}, no need to save lookup', title, series.name
        )
        return
    session.add(TVMazeLookup(search_name=title, series=series))


def prepare_lookup_for_tvmaze(**lookup_params):
    """
    Return a dict of params which is valid with tvmaze API lookups

    :param lookup_params: Search parameters
    :return: Dict of tvmaze recognizable key words
    """
    prepared_params = {}
    title = None
    series_name = (
        lookup_params.get('series_name')
        or lookup_params.get('show_name')
        or lookup_params.get('title')
    )
    if series_name:
        title, _ = split_title_year(series_name)
    # Support for when title is just a number
    if not title:
        title = series_name

    # Ensure we send native types to tvmaze lib as it does not handle new types very well
    prepared_params['tvmaze_id'] = lookup_params.get('tvmaze_id')
    prepared_params['thetvdb_id'] = lookup_params.get('tvdb_id') or lookup_params.get(
        'trakt_series_tvdb_id'
    )
    prepared_params['tvrage_id'] = lookup_params.get('tvrage_id') or lookup_params.get(
        'trakt_series_tvrage_id'
    )
    prepared_params['imdb_id'] = lookup_params.get('imdb_id')
    prepared_params['show_name'] = title or None

    return prepared_params


class APITVMaze:
    @staticmethod
    @with_session
    def series_lookup(session=None, only_cached=False, **lookup_params):
        search_params = search_params_for_series(**lookup_params)
        # Searching cache first
        series = from_cache(session=session, cache_type=TVMazeSeries, search_params=search_params)

        search = None
        # Preparing search from lookup table
        title = (
            lookup_params.get('series_name')
            or lookup_params.get('show_name')
            or lookup_params.get('title')
        )
        if not series and title:
            logger.debug(
                'did not find exact match for series {} in cache, looking in search table',
                search_params['name'],
            )
            search = from_lookup(session=session, title=title)
            if search and search.series:
                series = search.series
                logger.debug('found series {} from search table', series.name)

        if only_cached:
            if series:  # If force_cache is True, return series even if it expired
                logger.debug('forcing cache for series {}', series.name)
                return series
            raise LookupError('Series %s not found from cache' % lookup_params)
        if series and not series.expired:
            logger.debug('returning series {} from cache', series.name)
            return series

        prepared_params = prepare_lookup_for_tvmaze(**lookup_params)
        logger.debug('trying to fetch series {} from tvmaze', title)
        tvmaze_show = get_show(**prepared_params)

        # See if series already exist in cache
        series = (
            session.query(TVMazeSeries).filter(TVMazeSeries.tvmaze_id == tvmaze_show['id']).first()
        )
        if series:
            logger.debug('series {} is already in cache, checking for expiration', series.name)
            if series.expired:
                series.update(tvmaze_show, session)
        else:
            logger.debug('creating new series {} in tvmaze_series db', tvmaze_show['name'])
            series = TVMazeSeries(tvmaze_show, session)
            session.add(series)

        # Check if show returned from lookup table as expired. Relevant only if search by title
        if title:
            if series and title.lower() == series.name.lower():
                return series
            elif series and not search:
                logger.debug(
                    'mismatch between search title {} and series title {}. '
                    'saving in lookup table',
                    title,
                    series.name,
                )
                add_to_lookup(session=session, title=title, series=series)
            elif series and search:
                logger.debug('Updating search result in db')
                search.series = series
        return series

    @staticmethod
    @with_session
    def season_lookup(session=None, only_cached=False, **lookup_params):
        series_name = lookup_params.get('series_name') or lookup_params.get('title')
        show_id = lookup_params.get('tvmaze_id') or lookup_params.get('tvdb_id')

        season_number = lookup_params.get('series_season')

        # Verify we have enough parameters for search
        if not any([series_name, show_id, season_number]):
            raise LookupError('Not enough parameters to lookup episode')

        # Get series
        series = APITVMaze.series_lookup(session=session, only_cached=only_cached, **lookup_params)
        if not series:
            raise LookupError(
                'Could not find series with the following parameters: {0}'.format(lookup_params)
            )
        session.flush()
        # See if season already exists in cache
        logger.debug('searching for season {} of show {} in cache', season_number, series.name)
        season = (
            session.query(TVMazeSeason)
            .filter(TVMazeSeason.series_id == series.tvmaze_id)
            .filter(TVMazeSeason.number == season_number)
            .one_or_none()
        )

        # Logic for cache only mode
        if only_cached:
            if season:
                logger.debug('forcing cache for season {} of show {}', season_number, series.name)
                return season

        if season and not series.expired:
            logger.debug('returning season {} of show {}', season_number, series.name)
            return season

        # If no season has been found try refreshing the series seasons
        series.populate_seasons()

        # Query again
        season = (
            session.query(TVMazeSeason)
            .filter(TVMazeSeason.tvmaze_id == series.tvmaze_id)
            .filter(TVMazeSeason.number == season_number)
            .one_or_none()
        )
        if season:
            return season

    @staticmethod
    @with_session
    def episode_lookup(session=None, only_cached=False, **lookup_params):
        series_name = lookup_params.get('series_name') or lookup_params.get('title')
        show_id = lookup_params.get('tvmaze_id') or lookup_params.get('tvdb_id')
        lookup_type = lookup_params.get('series_id_type')

        season_number = lookup_params.get('series_season')
        episode_number = lookup_params.get('series_episode')

        episode_date = lookup_params.get('series_date')

        # Verify we have enough parameters for search
        if not any([series_name, show_id]):
            raise LookupError('Not enough parameters to lookup episode')
        if lookup_type == 'sequence':
            raise LookupError('TVMaze does not support sequence type searches')
        if lookup_type == 'ep' and not all([season_number, episode_number]):
            raise LookupError('Not enough parameters to lookup episode')
        elif lookup_type == 'date' and not episode_date:
            raise LookupError('Not enough parameters to lookup episode')

        # Get series
        series = APITVMaze.series_lookup(session=session, only_cached=only_cached, **lookup_params)
        if not series:
            raise LookupError(
                'Could not find series with the following parameters: {0}'.format(lookup_params)
            )

        # See if episode already exists in cache
        logger.debug('searching for episode of show {} in cache', series.name)
        episode = (
            session.query(TVMazeEpisodes)
            .filter(
                and_(
                    TVMazeEpisodes.series_id == series.tvmaze_id,
                    TVMazeEpisodes.season_number == season_number,
                    TVMazeEpisodes.number == episode_number,
                )
            )
            .one_or_none()
        )

        # Logic for cache only mode
        if only_cached:
            if episode:
                logger.debug(
                    'forcing cache for episode id {3}, number{0}, season {1} for show {2}',
                    episode.number,
                    episode.season_number,
                    series.name,
                    episode.tvmaze_id,
                )
                return episode
        if episode and not episode.expired:
            logger.debug(
                'found episode id {3}, number {0}, season {1} for show {2} in cache',
                episode.number,
                episode.season_number,
                series.name,
                episode.tvmaze_id,
            )

            return episode

        # Lookup episode via its type (number or airdate)
        if lookup_type == 'date':
            episode_date = datetime.strftime(episode_date, '%Y-%m-%d')
            tvmaze_episode = get_episode(series.tvmaze_id, date=episode_date)[0]
        else:
            # TODO will this match all series_id types?
            logger.debug(
                'fetching episode {0} season {1} for series_id {2} for tvmaze',
                episode_number,
                season_number,
                series.tvmaze_id,
            )
            tvmaze_episode = get_episode(
                series.tvmaze_id, season=season_number, number=episode_number
            )
        # See if episode exists in DB
        try:
            episode = (
                session.query(TVMazeEpisodes)
                .filter(
                    or_(
                        TVMazeEpisodes.tvmaze_id == tvmaze_episode['id'],
                        and_(
                            TVMazeEpisodes.number == tvmaze_episode['number'],
                            TVMazeEpisodes.season_number == tvmaze_episode['season'],
                            TVMazeEpisodes.series_id == series.tvmaze_id,
                        ),
                    )
                )
                .one_or_none()
            )
        except MultipleResultsFound:
            # TVMaze must have fucked up and now we have to clean up that mess. Delete any row for this season
            # that hasn't been updated in the last hour. Can't trust any of the cached data, but deleting new data
            # might have some unintended consequences.
            logger.warning(
                'Episode lookup in cache returned multiple results. Deleting the cached data.'
            )
            deleted_rows = (
                session.query(TVMazeEpisodes)
                .filter(
                    and_(
                        TVMazeEpisodes.season_number == tvmaze_episode['season'],
                        TVMazeEpisodes.series_id == series.tvmaze_id,
                    )
                )
                .filter(TVMazeEpisodes.last_update <= datetime.now() - timedelta(hours=1))
                .delete()
            )
            logger.debug('Deleted {} rows', deleted_rows)
            episode = None

        if episode:
            logger.debug('found expired episode {} in cache, refreshing data.', episode.tvmaze_id)
            episode.update(tvmaze_episode)
        else:
            logger.debug('creating new episode for show {}', series.name)
            episode = TVMazeEpisodes(tvmaze_episode, series.tvmaze_id)
            session.add(episode)

        return episode


def get_show(show_name=None, tvmaze_id=None, imdb_id=None, tvrage_id=None, thetvdb_id=None):
    if not any(param for param in [show_name, tvmaze_id, imdb_id, tvrage_id, thetvdb_id]):
        raise LookupError('Not enough parameters sent for series lookup')
    params = {'embed': 'seasons'}
    if tvmaze_id:
        url = TVMAZE_SHOW_PATH.format(tvmaze_id)
    elif imdb_id:
        url = TVMAZE_LOOKUP_PATH
        params['imdb'] = imdb_id
    elif tvrage_id:
        url = TVMAZE_LOOKUP_PATH
        params['tvrage'] = tvrage_id
    elif thetvdb_id:
        url = TVMAZE_LOOKUP_PATH
        params['thetvdb'] = thetvdb_id
    else:
        params['q'] = show_name
        url = TVMAZE_SEARCH_PATH
    return tvmaze_lookup(url, params=params)


def get_episode(series_id, date=None, number=None, season=None):
    if date:
        return tvmaze_lookup(TVMAZE_EPISODES_BY_DATE_PATH.format(series_id), params={'date': date})
    elif number and season:
        return tvmaze_lookup(
            TVMAZE_EPISODES_BY_NUMBER_PATH.format(series_id),
            params={'season': season, 'number': number},
        )
    raise LookupError('Not enough parameters sent for episode lookup')


def get_seasons(series_id):
    return tvmaze_lookup(TVMAZE_SEASONS.format(series_id))


def tvmaze_lookup(lookup_url, **kwargs):
    """
    Build the URL and return the reply from TVMaze API

    :param lookup_type: Selects the endpoint that will be used
    :param lookup_values: A list of values to be used in the URL
    :return: A JSON reply from the API
    """
    url = BASE_URL + lookup_url
    logger.debug('querying tvmaze API with the following URL: {}', url)
    try:
        result = requests.get(url, **kwargs).json()
    except RequestException as e:
        raise LookupError(e.args[0])
    return result


@event('plugin.register')
def register_plugin():
    plugin.register(APITVMaze, 'api_tvmaze', api_ver=2, interfaces=[])



"""
Collection of generic parser related utilities used by internal parser and with
parsing plugins.
"""
import re


class ParseWarning(Warning):
    def __init__(self, parsed, value, **kwargs):
        self.value = value
        self.parsed = parsed
        self.kwargs = kwargs

    def __unicode__(self):
        return self.value

    def __str__(self):
        return self.__unicode__().encode('utf-8')

    def __repr__(self):
        return str('ParseWarning({}, **{})').format(self, repr(self.kwargs))


default_ignore_prefixes = [
    r'(?:\[[^\[\]]*\])',  # ignores group names before the name, eg [foobar] name
    r'(?:HD.720p?:)',
    r'(?:HD.1080p?:)',
    r'(?:HD.2160p?:)',
]


def name_to_re(name, ignore_prefixes=None, parser=None):
    """Convert 'foo bar' to '^[^...]*foo[^...]*bar[^...]+"""
    if not ignore_prefixes:
        ignore_prefixes = default_ignore_prefixes
    parenthetical = None
    if name.endswith(')'):
        p_start = name.rfind('(')
        if p_start != -1:
            parenthetical = re.escape(name[p_start + 1 : -1])
            name = name[: p_start - 1]
    # Blanks are any non word characters except & and _
    blank = r'(?:[^\w&]|_)'
    ignore = '(?:' + '|'.join(ignore_prefixes) + ')?'
    res = re.sub(re.compile(blank + '+', re.UNICODE), ' ', name)
    res = res.strip()
    # accept either '&' or 'and'
    res = re.sub(' (&|and) ', ' (?:and|&) ', res, re.UNICODE)
    # The replacement has a regex escape in it (\w) which needs to be escaped again in python 3.7+
    res = re.sub(' +', blank.replace('\\', '\\\\') + '*', res, re.UNICODE)
    if parenthetical:
        res += '(?:' + blank + '+' + parenthetical + ')?'
        # Turn on exact mode for series ending with a parenthetical,
        # so that 'Show (US)' is not accepted as 'Show (UK)'
        if parser:
            parser.strict_name = True
    res = '^' + ignore + blank + '*' + '(' + res + ')(?:\\b|_)' + blank + '*'
    return res



import re

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='modify_torrents')


class AddTrackers:
    """
    Adds tracker URL to torrent files.

    Configuration example:

    add_trackers:
      - uri://tracker_address:port/

    This will add all tracker URL uri://tracker_address:port/.
    """

    schema = {'type': 'array', 'items': {'type': 'string', 'format': 'url'}}

    @plugin.priority(127)
    def on_task_modify(self, task, config):
        for entry in task.entries:
            if 'torrent' in entry:
                for url in config:
                    if url not in entry['torrent'].trackers:
                        entry['torrent'].add_multitracker(url)
                        logger.info('Added {} tracker to {}', url, entry['title'])
            if entry['url'].startswith('magnet:'):
                entry['url'] += ''.join(['&tr=' + url for url in config])


class RemoveTrackers:
    """
    Removes trackers from torrent files using regexp matching.

    Configuration example:

    remove_trackers:
      - moviex

    This will remove all trackers that contain text moviex in their url.
    """

    schema = {'type': 'array', 'items': {'type': 'string', 'format': 'regex'}}

    @plugin.priority(127)
    def on_task_modify(self, task, config):
        for entry in task.entries:
            if 'torrent' in entry:
                for tracker in entry['torrent'].trackers:
                    for regexp in config or []:
                        if re.search(regexp, tracker, re.IGNORECASE | re.UNICODE):
                            logger.debug(
                                'remove_trackers removing {} because of {}', tracker, regexp
                            )
                            # remove tracker
                            entry['torrent'].remove_multitracker(tracker)
                            logger.info('Removed {}', tracker)
            if entry['url'].startswith('magnet:'):
                for regexp in config:
                    # Replace any tracker strings that match the regexp with nothing
                    tr_search = r'&tr=([^&]*%s[^&]*)' % regexp
                    entry['url'] = re.sub(tr_search, '', entry['url'], re.IGNORECASE | re.UNICODE)


class ModifyTrackers:
    """
    Modify tracker URL to torrent files.

    Configuration example::

        modify_trackers:
          - SearchAndReplace:
              from: string_to_search
              to: string_to_replace

    """

    schema = {
        'type': 'array',
        'items': {
            'type': 'object',
            'additionalProperties': {
                'type': 'object',
                'properties': {'from': {'type': 'string'}, 'to': {'type': 'string'}},
                'additionalProperties': False,
            },
            'maxProperties': 1,
        },
    }

    @plugin.priority(127)
    def on_task_modify(self, task, config):
        for entry in task.entries:
            if 'torrent' in entry:
                torrent = entry['torrent']
                trackers = torrent.trackers
                for item in config:
                    for replace in item.values():
                        for tracker in trackers:
                            if replace.get('from') in tracker:
                                torrent.remove_multitracker(tracker)
                                trackernew = tracker.replace(
                                    replace.get('from'), replace.get('to')
                                )
                                torrent.add_multitracker(trackernew)
                                logger.info('Modify {} in {}', tracker, trackernew)


@event('plugin.register')
def register_plugin():
    plugin.register(AddTrackers, 'add_trackers', api_ver=2)
    plugin.register(RemoveTrackers, 'remove_trackers', api_ver=2)
    plugin.register(ModifyTrackers, 'modify_trackers', api_ver=2)



from argparse import ArgumentParser, ArgumentTypeError

from sqlalchemy.orm.exc import NoResultFound

from flexget import options, plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.manager import Session
from flexget.plugin import DependencyError, PluginError
from flexget.terminal import TerminalTable, TerminalTableError, console, table_parser
from flexget.utils.tools import split_title_year

from . import db
from .movie_list import MovieListBase


def lookup_movie(title, session, identifiers=None):
    try:
        imdb_lookup = plugin.get('imdb_lookup', 'movie_list').lookup
    except DependencyError:
        imdb_lookup = None

    try:
        tmdb_lookup = plugin.get('tmdb_lookup', 'movie_list').lookup
    except DependencyError:
        tmdb_lookup = None

    if not (imdb_lookup or tmdb_lookup):
        return

    entry = Entry(title=title)
    if identifiers:
        for identifier in identifiers:
            for key, value in identifier.items():
                entry[key] = value
    try:
        imdb_lookup(entry, session=session)
    # IMDB lookup raises PluginError instead of the normal ValueError
    except PluginError:
        tmdb_lookup(entry)

    # Return only if lookup was successful
    if entry.get('movie_name'):
        return entry
    return


def movie_list_keyword_type(identifier):
    if identifier.count('=') != 1:
        raise ArgumentTypeError(
            'Received identifier in wrong format: {}, '
            ' should be in keyword format like `imdb_id=tt1234567`'.format(identifier)
        )
    name, value = identifier.split('=', 2)
    if name not in MovieListBase().supported_ids:
        raise ArgumentTypeError(
            'Received unsupported identifier ID {}. Should be one of {}'.format(
                identifier, ' ,'.join(MovieListBase().supported_ids)
            )
        )
    return {name: value}


def do_cli(manager, options):
    """Handle movie-list subcommand"""
    if options.list_action == 'all':
        movie_list_lists(options)
        return

    if options.list_action == 'list':
        movie_list_list(options)
        return

    if options.list_action == 'add':
        movie_list_add(options)
        return

    if options.list_action == 'del':
        movie_list_del(options)
        return

    if options.list_action == 'purge':
        movie_list_purge(options)
        return


def movie_list_lists(options):
    """ Show all movie lists """
    lists = db.get_movie_lists()
    header = ['#', 'List Name']
    table_data = [header]
    for movie_list in lists:
        table_data.append([movie_list.id, movie_list.name])
    try:
        table = TerminalTable(options.table_type, table_data)
    except TerminalTableError as e:
        console('ERROR: {}'.format(e))
    else:
        console(table.output)


def movie_list_list(options):
    """List movie list"""
    with Session() as session:
        try:
            movie_list = db.get_list_by_exact_name(options.list_name)
        except NoResultFound:
            console('Could not find movie list with name {}'.format(options.list_name))
            return
    header = ['#', 'Movie Name', 'Movie year']
    header += db.MovieListBase().supported_ids
    table_data = [header]
    movies = db.get_movies_by_list_id(
        movie_list.id, order_by='added', descending=True, session=session
    )
    for movie in movies:
        movie_row = [movie.id, movie.title, movie.year or '']
        for identifier in db.MovieListBase().supported_ids:
            movie_row.append(movie.identifiers.get(identifier, ''))
        table_data.append(movie_row)
    title = '{} Movies in movie list: `{}`'.format(len(movies), options.list_name)
    try:
        table = TerminalTable(options.table_type, table_data, title, drop_columns=[5, 2, 4])
    except TerminalTableError as e:
        console('ERROR: {}'.format(e))
    else:
        console(table.output)


def movie_list_add(options):
    with Session() as session:
        try:
            movie_list = db.get_list_by_exact_name(options.list_name, session=session)
        except NoResultFound:
            console('Could not find movie list with name {}, creating'.format(options.list_name))
            movie_list = db.MovieListList(name=options.list_name)
            session.add(movie_list)
            session.commit()

        title, year = split_title_year(options.movie_title)
        console('Trying to lookup movie title: `{}`'.format(title))
        movie_lookup = lookup_movie(title=title, session=session, identifiers=options.identifiers)
        if not movie_lookup:
            console(
                'ERROR: movie lookup failed for movie {}, aborting'.format(options.movie_title)
            )
            return

        title = movie_lookup['movie_name']
        movie = db.get_movie_by_title_and_year(
            list_id=movie_list.id, title=title, year=year, session=session
        )
        if not movie:
            console("Adding movie with title {} to list {}".format(title, movie_list.name))
            movie = db.MovieListMovie(title=title, year=year, list_id=movie_list.id)
        else:
            console("Movie with title {} already exist in list {}".format(title, movie_list.name))

        id_list = []
        if options.identifiers:
            id_list = options.identifiers
        else:
            for _id in db.MovieListBase().supported_ids:
                if movie_lookup.get(_id):
                    id_list.append({_id: movie_lookup.get(_id)})
        if id_list:
            console('Setting movie identifiers:')
            for ident in id_list:
                for key in ident:
                    console('{}: {}'.format(key, ident[key]))
            movie.ids = db.get_db_movie_identifiers(identifier_list=id_list, session=session)
        session.merge(movie)
        console('Successfully added movie {} to movie list {} '.format(title, movie_list.name))


def movie_list_del(options):
    with Session() as session:
        try:
            movie_list = db.get_list_by_exact_name(options.list_name)
        except NoResultFound:
            console('Could not find movie list with name {}'.format(options.list_name))
            return

        try:
            movie_exist = db.get_movie_by_id(
                list_id=movie_list.id, movie_id=int(options.movie), session=session
            )
        except NoResultFound:
            console(
                'Could not find movie with ID {} in list `{}`'.format(
                    int(options.movie), options.list_name
                )
            )
            return
        except ValueError:
            title, year = split_title_year(options.movie)
            movie_exist = db.get_movie_by_title_and_year(
                list_id=movie_list.id, title=title, year=year, session=session
            )
        if not movie_exist:
            console(
                'Could not find movie with title {} in list {}'.format(
                    options.movie, options.list_name
                )
            )
            return
        else:
            console('Removing movie {} from list {}'.format(movie_exist.title, options.list_name))
            session.delete(movie_exist)


def movie_list_purge(options):
    with Session() as session:
        try:
            movie_list = db.get_list_by_exact_name(options.list_name)
        except NoResultFound:
            console('Could not find movie list with name {}'.format(options.list_name))
            return
        console('Deleting list {}'.format(options.list_name))
        session.delete(movie_list)


@event('options.register')
def register_parser_arguments():
    # Common option to be used in multiple subparsers
    movie_parser = ArgumentParser(add_help=False)
    movie_parser.add_argument('movie_title', metavar='<MOVIE TITLE>', help="Title of the movie")

    name_or_id_parser = ArgumentParser(add_help=False)
    name_or_id_parser.add_argument(
        'movie', metavar='<NAME or ID>', help="Title or ID of the movie"
    )

    identifiers_parser = ArgumentParser(add_help=False)
    identifiers_parser.add_argument(
        '-i',
        '--identifiers',
        metavar='<identifiers>',
        nargs='+',
        type=movie_list_keyword_type,
        help='Can be a string or a list of string with the format imdb_id=XXX,'
        ' tmdb_id=XXX, etc',
    )
    list_name_parser = ArgumentParser(add_help=False)
    list_name_parser.add_argument(
        'list_name',
        nargs='?',
        metavar='<LIST NAME>',
        default='movies',
        help='Name of movie list to operate on (Default is \'movies\')',
    )
    # Register subcommand
    parser = options.register_command('movie-list', do_cli, help='View and manage movie lists')
    # Set up our subparsers
    subparsers = parser.add_subparsers(title='actions', metavar='<action>', dest='list_action')
    subparsers.add_parser('all', parents=[table_parser], help='Shows all existing movie lists')
    subparsers.add_parser(
        'list', parents=[list_name_parser, table_parser], help='List movies from a list'
    )
    subparsers.add_parser(
        'add',
        parents=[list_name_parser, movie_parser, identifiers_parser],
        help='Add a movie to a list',
    )
    subparsers.add_parser(
        'del',
        parents=[list_name_parser, name_or_id_parser],
        help='Remove a movie from a list using its title or ID',
    )
    subparsers.add_parser(
        'purge',
        parents=[list_name_parser],
        help='Removes an entire list with all of its movies. Use this with caution',
    )
